###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################

from numpy import array, sum, sqrt, log, abs, loadtxt, zeros_like, shape
from matplotlib.pyplot import plot, figure, subplot, show, ion, title, suptitle, legend, gca, ioff, axvline, gcf
import subprocess, sys, os, time

print()
print('  ########  start script rms_tracker  ########')
print('  please give input path and some options:')
print('  available options are: 0 = rms')
print('                         1 = total energy')
print('                         2 = charge neutrality')
print('                         3 = total moment')
print()
print('  if last given option is -1 then no refreshing is done and plots are only generated once')
print()
print(
    "  after the options there is the possibility to give the 'nodos' option which prevents the plotting of the dos if 'dos.atom*' files are there"
)
print()
print('  input might look like this:')
print('    for self refreshing plot of rms charge neutrality and moment: ./rms_tracker.py ./path/ 0 2 3')
print('    for non refreshing plot of rms and total energy without dos plot: ./rms_tracker.py ./path/ 0 1 -1 nodos')
print()
print('  default values: ./rms_tracker.py ./path/ 0 1 2 3')

if len(sys.argv) > 1:
    allpaths = sys.argv[1]
    # create list of paths from input
    allpath = []
    for path0 in allpaths.split(','):
        outfile_found = False
        for file in ['out', 'out_kkr']:
            if file in os.listdir(path0):
                outfile_found = True
                outfile = file
        if outfile_found:
            if path0 == '.' or path0 == './':
                path0 = os.getcwd()
            if path0[-1] != '/':
                path0 += '/'
            allpath.append(path0)
        else:
            print("WARNING: file 'out' not found in", path0)
    if allpath == []:
        print("no file 'out' found")
        sys.exit()

    if len(sys.argv) > 2:
        nopt = len(sys.argv[2:])
        opt = sys.argv[2:]
        if len(opt[-1]) <= 2:
            dos = 'dos.atom1' in os.listdir(path0)
        else:
            dos = False
            nopt = nopt - 1
            opt = opt[:-1]
        opt = [int(i) for i in opt]
        print(opt)
        if opt[-1] == -1:
            refresh = False
            nopt = nopt - 1
            opt = opt[:-1]
        else:
            refresh = True
    else:
        nopt = 4
        opt = [0, 1, 2, 3]
        refresh = True
        if len(sys.argv[-1]) <= 2:
            dos = 'dos.atom1' in os.listdir(path0)
        else:
            dos = False

else:
    print('Please give input path')
    sys.exit()

# consistency check and names
names = []
for i in opt:
    if i not in [0, 1, 2, 3]:
        print('Found non existing option:', i)
        sys.exit()
    else:
        if i == 0:
            names.append('rms error')
        elif i == 1:
            names.append('total energy')
        elif i == 2:
            names.append('neutrality')
        elif i == 3:
            names.append('total moment')

# print input
print()
print('  your input options:', opt, '; refresh=', refresh, '; DOS=', dos)
print('  path(s):', allpath)
print()


def read_rms_data(path0):
    global opt, nopt
    ### rms
    try:
        f = subprocess.check_output('grep aver ' + path0 + outfile, shell=True).decode('utf-8').split('\n')
        rms = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        rms = []

    ### charge neutrality
    try:
        f = subprocess.check_output('grep neutr ' + path0 + outfile, shell=True).decode('utf-8').split('\n')
        neut = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        neut = []
    # check if neutrality info is available (not there in case of impurity code output)
    if len(neut) < len(rms) and 2 in opt:
        nopt = nopt - 1
        j = 0
        for i in range(nopt + 1):
            if opt[i] == 2:
                j = i
        print('removing option', opt.pop(j), names.pop(j))

    ### read total energy
    try:
        f = subprocess.check_output("grep 'TOTAL ENERGY' " + path0 + outfile, shell=True).decode('utf-8').split('\n')
        etot = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        etot = []

    ### moment
    try:
        f = subprocess.check_output("grep 'L m' " + path0 + outfile, shell=True).decode('utf-8').split('\n')
        mom = array([float((i.split()[-1]).replace('D', 'e').replace('\n', '')) for i in f if i != ''])
    except:
        mom = []
    # check if moment info is available
    if len(mom) < len(rms) and 3 in opt:
        nopt = nopt - 1
        for i in range(nopt + 1):
            if opt[i] == 3:
                j = i
        print('removing option', opt.pop(j), names.pop(j))
    else:
        try:
            f = subprocess.check_output("grep 'ITERATION ' " + path0 + outfile, shell=True).decode('utf-8').split('\n')
            tmp = array(
                [float((i.split()[1]).replace('D', 'e').replace('\n', '')) for i in f if 'SCF' not in i and i != ''])
        except:
            tmp = []
        if tmp != [] and 'out_magneticmoments' in os.listdir(path0):
            it = int(tmp[-1])
            tmp = loadtxt(path0 + 'out_magneticmoments')
            ncls = len(tmp) // it
            tmp = tmp[::ncls, :3]
            tmpmomx = tmp[:, 0]
            tmpmomy = tmp[:, 1]
            tmpmomz = tmp[:, 2]
            tmpmom = mom
            mom = array([tmpmom, tmpmomx, tmpmomy, tmpmomz])
        else:
            inp = open(path0 + 'inputcard').readlines()
            natyp = -1
            for iline in range(len(inp)):
                if 'NATYP' in inp[iline]:
                    natyp = iline
            if natyp == -1:
                for iline in range(len(inp)):
                    if 'NAEZ' in inp[iline]:
                        natyp = iline
                natyp = int(inp[natyp].split('NAEZ=')[1])
            else:
                natyp = int(inp[natyp].split('NATYP=')[1])
            try:
                tmp2 = subprocess.check_output("grep 'm_spin' " + path0 + outfile + ' -A' + str(natyp),
                                               shell=True).decode('utf-8').replace('TOT', '').split('\n')
                if tmp2 != '':
                    tmp2 = tmp2  #[-natyp-1:]
                    tmp2 = array([float(i.split()[2]) for i in tmp2 if i != '' and 'dn' not in i and i != '--'])
                    tmp2 = tmp2.reshape(-1, natyp)
                    mom = array([mom] + [tmp2[:, i] for i in range(len(tmp2[0]))])
            except:
                tmpmom, tmpmomx, tmpmomy, tmpmomz = mom, mom, mom, mom
                mom = array([tmpmom, tmpmomx, tmpmomy, tmpmomz])

    return rms, neut, mom, etot


# turn interactive mode on (needed for figures to appear immediately in while loop)
ion()

rms = [[] for i in range(len(allpath))]
neut = [[] for i in range(len(allpath))]
etot = [[] for i in range(len(allpath))]
mom = [[] for i in range(len(allpath))]
istart = [True for i in range(len(allpath))]

while (True):
    for ipath0 in range(len(allpath)):
        path0 = allpath[ipath0]
        l0 = len(rms[ipath0])
        rms[ipath0], neut[ipath0], mom[ipath0], etot[ipath0] = read_rms_data(path0)

        if len(rms[ipath0]) > l0:
            ### fig 1 for plots of values from option
            figure(ipath0)
            figure(ipath0).clf()

            # give path name to figure
            gcf().canvas.set_window_title(allpath[ipath0])

            for i in range(nopt):
                # choose data according to option
                if opt[i] == 0:
                    data = rms[ipath0]
                elif opt[i] == 1:
                    data = etot[ipath0]
                elif opt[i] == 2:
                    data = neut[ipath0]
                elif opt[i] == 3:
                    data = mom[ipath0]

                # choose appropriate subplot according to number of options
                if nopt == 1:
                    subplot(1, 2, 1 + i)
                elif nopt == 2:
                    subplot(2, 2, 1 + i)
                elif nopt == 3:
                    subplot(2, 3, 1 + i)
                elif nopt == 4:
                    subplot(2, 4, 1 + i)

                # plot data in linear scale
                title(names[i])
                if opt[i] == 3:
                    clrs = ['b', 'g', 'r', 'y', 'm', 'c', 'k']
                    for j in range(len(data) - 1, -1, -1):
                        plot(data[j], 'x--')  #+clrs[j])
                else:
                    plot(data, 'x--')

                # subtract reference for Etot and mom
                if len(data) > 1 and opt[i] in [1, 3]:
                    if opt[i] == 3 and len(data[0]) > 1:
                        datanew = zeros_like(data)[:, :-1]
                        for j in range(len(data[:, 0])):
                            datanew[j, :] = (data[j, :] - data[j, -1])[:-1]
                        data = datanew
                    else:
                        data = (data - data[-1])[:-1]

                #  plot log scale if data is not always 0
                if (abs(data).any() > 0):
                    if nopt == 1:
                        subplot(1, 2, 1 + i + nopt)
                    elif nopt == 2:
                        subplot(2, 2, 1 + i + nopt)
                    elif nopt == 3:
                        subplot(2, 3, 1 + i + nopt)
                    elif nopt == 4:
                        subplot(2, 4, 1 + i + nopt)
                    if opt[i] in [1, 3]:
                        name = 'change in ' + names[i]
                    else:
                        name = names[i] + ', log scale'
                    title(name)
                    if opt[i] == 3:
                        for j in range(len(data) - 1, -1, -1):
                            plot(abs(data[j]), 'x--')  #+clrs[j])
                    else:
                        plot(abs(data), 'x--')
                    gca().set_yscale('log')
                try:
                    t_iter = subprocess.check_output('grep Iter ' + path0 + 'out_timing.000.txt',
                                                     shell=True).decode('utf-8').split('\n')[-2].split()[-1]
                    t_iter = f'{float(t_iter) / 60:5.2f} min'
                except:
                    t_iter = 'no time info for last iteration'
                suptitle(time.ctime() + ', time in last iteration: ' + t_iter)

            ### fig 2 for DOS plots
            if dos:
                figure(ipath0 + 1000)
                figure(ipath0 + 1000).clf()
                j = 0
                for i in os.listdir(path0):
                    if len(i) > 8 and 'dos.atom' == i[:8]:
                        tmpi = int(i.replace('dos.atom', ''))
                        if j <= tmpi:
                            j = tmpi

                if len(rms[ipath0]) > 1 and not istart[ipath0]:
                    subplot(1, 2, 2)
                    title('previous iteration')
                    for i in range(len(d_old)):
                        d = d_old[i]
                        lab = i
                        plot(d[:, 0], sum(d[:, 1:], axis=1), '-', label=lab)
                    legend()
                    ef = float(open(path0 + 'potential').readlines()[3].split()[1])
                    axvline(ef)
                    subplot(1, 2, 1)
                else:
                    subplot(1, 1, 1)
                title('current iteration')
                d_old = []
                for i in range(j):
                    d = loadtxt(path0 + 'dos.atom%i' % (i + 1))
                    lab = i
                    plot(d[:, 0], sum(d[:, 1:], axis=1), '-', label=lab)
                    d_old.append(d)
                    istart[ipath0] = False
                legend()
                ef = float(open(path0 + 'potential').readlines()[3].split()[1])
                axvline(ef)
                suptitle(time.ctime())

        figure(ipath0).canvas.draw()
        figure(ipath0).canvas.flush_events()

        if dos:
            figure(ipath0 + 1000).canvas.draw()
            figure(ipath0 + 1000).canvas.flush_events()

        dobreak = False
        if refresh:
            time.sleep(2)
        else:
            # turn off interactive mode and exit while loop
            if path0 == allpath[-1]:
                dobreak = True
    if dobreak:
        break

ioff()
show()



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains functions to load an fleur out.xml file, parse it with a schema
and convert its content to a dict, based on the tasks given
"""
from __future__ import annotations
from functools import partial

from masci_tools.util.xml.common_functions import clear_xml
from masci_tools.util.xml.converters import convert_str_version_number
from masci_tools.util.xml import xml_getters
from masci_tools.util.xml.xpathbuilder import FilterType
from masci_tools.util.parse_utils import Conversion
from masci_tools.io.fleur_xml import FleurXMLContext, load_outxml_and_check_for_broken_xml, _EvalContext
from masci_tools.util.logging_util import DictHandler, OutParserLogAdapter
from masci_tools.util.typing import XMLFileLike
import copy
import warnings
import logging
from typing import Any, Callable, Iterable, TypeVar
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  #type:ignore
import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

__all__ = ('outxml_parser', 'conversion_function', 'register_migration')


def outxml_parser(outxmlfile: XMLFileLike,
                  parser_info_out: dict[str, Any] | None = None,
                  iteration_to_parse: Literal['all', 'last', 'first'] | int = 'last',
                  minimal_mode: bool = False,
                  additional_tasks: dict[str, dict[str, Any]] | None = None,
                  optional_tasks: Iterable[str] | None = None,
                  overwrite: bool = False,
                  append: bool = False,
                  list_return: bool = False,
                  strict: bool = False,
                  debug: bool = False,
                  ignore_validation: bool = False,
                  base_url: str | None = None) -> dict[str, Any]:
    """
    Parses the out.xml file to a dictionary based on the version and the given tasks

    :param outxmlfile: either path to the out.xml file, opened file handle (in bytes modes i.e. rb)
                       or a xml etree to be parsed
    :param parser_info_out: dict, with warnings, info, errors, ...
    :param iteration_to_parse: either str or int, (optional, default 'last')
                               determines which iteration should be parsed.
                               Accepted are 'all', 'first', 'last' or an index for the iteration
    :param minimal_mode: bool, if True only total Energy, iteration number and distances are parsed
    :param additional_tasks: dict to define custom parsing tasks. For detailed explanation
                             See :py:mod:`~masci_tools.io.parsers.fleur.default_parse_tasks`.
    :param overwrite: bool, if True and keys in additional_tasks collide with defaults
                      The defaults will be overwritten
    :param append: bool, if True and keys in additional_tasks collide with defaults
                   The inner tasks will be written into the dict. If inner keys collide
                   they are overwritten
    :param optional_tasks: Iterable of strings, defines additional tasks to perform.
                           See :py:mod:`~masci_tools.io.parsers.fleur.default_parse_tasks` for examples.
    :param list_return: bool, if True one-item lists in the output dict are not converted to simple values
    :param strict: bool if True  and no parser_info_out is provided any encountered error will immediately be raised
    :param debug: bool if True additional information is printed out in the logs
    :param ignore_validation: bool, if True schema validation errors are only logged

    :return: python dictionary with the information parsed from the out.xml

    :raises ValueError: If the validation against the schema failed, or an irrecoverable error
                        occurred during parsing
    :raises FileNotFoundError: If no Schema file for the given version was found
    :raises KeyError: If an unknown task is encountered
    """

    __parser_version__ = '0.7.0'

    logger: logging.Logger | None = logging.getLogger(__name__)
    if strict:
        logger = None

    parser_log_handler = None
    if logger is not None:
        if parser_info_out is None:
            parser_info_out = {}

        logging_level = logging.INFO
        if debug:
            logging_level = logging.DEBUG
        logger.setLevel(logging_level)

        parser_log_handler = DictHandler(parser_info_out,
                                         WARNING='parser_warnings',
                                         ERROR='parser_errors',
                                         INFO='parser_info',
                                         DEBUG='parser_debug',
                                         CRITICAL='parser_critical',
                                         ignore_unknown_levels=True,
                                         level=logging_level)

        logger.addHandler(parser_log_handler)

    if logger is not None:
        logger.info('Masci-Tools Fleur out.xml Parser v%s', __parser_version__)

    try:
        xmltree, schema_dict, outfile_broken = load_outxml_and_check_for_broken_xml(outxmlfile,
                                                                                    logger=logger,
                                                                                    base_url=base_url)
    except ValueError as err:
        if logger is not None:
            logger.error(str(err))
        if 'Skipping the parsing of the XML file' in str(err):
            return {}
        raise
    xmltree, _ = clear_xml(xmltree)

    with FleurXMLContext(xmltree, schema_dict, logger=logger) as root:

        out_version = root.attribute('fleurOutputVersion')
        if out_version == '0.27':
            inp_version = out_version
        else:
            inp_version = root.attribute('fleurInputVersion')

        if schema_dict['out_version'] != out_version or \
           schema_dict['inp_version'] != inp_version:
            ignore_validation = True
            out_version = schema_dict['out_version']
            inp_version = schema_dict['inp_version']

        if logger is not None:
            logger.info('Found fleur out file with the versions out: %s; inp: %s', out_version, inp_version)

        try:
            schema_dict.validate(xmltree, logger=logger)
        except ValueError as err:
            if not ignore_validation:
                if logger is not None:
                    logger.exception(err)
                raise

        parser = _TaskParser(out_version)
        if additional_tasks is None:
            additional_tasks = {}
        for task_name, task_definition in additional_tasks.items():
            parser.add_task(task_name, task_definition, overwrite=overwrite, append=append)

        if logger is not None:
            logger.info('The following defined constants were found: %s', root.constants)

        fleur_modes = xml_getters.get_fleur_modes(xmltree, schema_dict, logger=logger)
        if logger is not None:
            logger.info('The following Fleur modes were found: %s', fleur_modes)
        parser.determine_tasks(fleur_modes, optional_tasks, minimal=minimal_mode, iteration_to_parse=iteration_to_parse)

        out_dict = {'input_file_version': schema_dict['inp_version'], 'fleur_modes': fleur_modes}
        if logger is not None:
            logger.debug('The following tasks are performed on the root: %s', parser.general_tasks)
        for task in parser.general_tasks:

            if logger is not None:
                logger.debug('Performing task: %s', task)
            out_dict = parser.perform_task(task, root, out_dict, use_lists=False)

        iteration_filter = _determine_iteration_condition(iteration_to_parse, root.number_nodes('iteration'),
                                                          outfile_broken, logger)

        logger_info: dict[str, Any] = {}
        iteration_logger: logging.LoggerAdapter | None = None
        if logger is not None:
            iteration_logger = OutParserLogAdapter(logger, logger_info)

        for iteration in root.iter('iteration', filters=iteration_filter):
            iteration.logger = iteration_logger  #type:ignore[assignment] #TODO: Should this be allowed to be overwritten in iter?
            logger_info['iteration'] = iteration.attribute('numberForCurrentRun', default='unknown')

            iteration_tasks = parser.iteration_tasks
            #If the iteration is a forcetheorem calculation
            #Replace all tasks with the given tasks for the calculation
            forcetheorem_tags = ['Forcetheorem_DMI', 'Forcetheorem_SSDISP', 'Forcetheorem_JIJ', 'Forcetheorem_MAE']
            for tag in forcetheorem_tags:
                if iteration.tag_exists(tag):
                    if minimal_mode:
                        iteration_tasks = []
                    else:
                        iteration_tasks = [tag.lower()]
                    break

            if iteration.logger is not None:
                iteration.logger.debug('The following tasks are performed for the iteration: %s', iteration_tasks)

            for task in iteration_tasks:

                if iteration.logger is not None:
                    iteration.logger.debug('Performing task: %s', task)

                try:
                    out_dict = parser.perform_task(task, iteration, out_dict)
                except KeyError:
                    if logger is not None:
                        logger.exception("Unknown task: '%s'. Skipping this one", task)
                    raise

    if not list_return:
        #Convert one item lists to simple values
        for key, value in out_dict.items():
            if isinstance(value, list):
                if len(value) == 1:
                    out_dict[key] = value[0]
            elif isinstance(value, dict):
                for subkey, subvalue in value.items():
                    if isinstance(subvalue, list):
                        if len(subvalue) == 1:
                            out_dict[key][subkey] = subvalue[0]

    if parser_log_handler is not None:
        if logger is not None:
            logger.removeHandler(parser_log_handler)

    return out_dict


def _determine_iteration_condition(iteration_to_parse: Literal['all', 'first', 'last'] | int, n_iters: int,
                                   broken: bool, logger: logging.Logger | None) -> FilterType:
    """
    Determine which iterations should be parsed
    If the XML file is broken the last stable iteration is parsed if possible

    :param iteration_to_parse: either str or int, (optional, default 'last')
                               determines which iteration should be parsed.
                               Accepted are 'all', 'first', 'last' or an index for the iteration
    :param n_iters: How many iterations are in the file
    :param broken: if True the last iteration is assumed to be broken
    :param logger: logger for warnings
    """
    if n_iters == 0:
        # there was no iteration found.
        # only the starting charge density could be generated
        msg = 'There was no iteration found in the outfile, either just a ' \
              'starting density was generated or something went wrong.'
        if logger is None:
            raise ValueError(msg)
        logger.error(msg)
        return {}

    if logger is not None and broken and n_iters == 1:
        logger.info('The last parsed iteration is %s', n_iters)

    index_condition: int | dict[str, int] | None = None
    if iteration_to_parse == 'last':
        index_condition = -1
        if broken and n_iters >= 2:
            index_condition = -2
            if logger is not None:
                logger.info('The last parsed iteration is %s', n_iters - 2)
    elif iteration_to_parse == 'first':
        index_condition = 1
    elif isinstance(iteration_to_parse, int):
        if iteration_to_parse >= n_iters or iteration_to_parse < -n_iters:
            if logger is not None:
                logger.error("Invalid value for iteration_to_parse: Got '%s'", iteration_to_parse)
            raise ValueError(f"Invalid value for iteration_to_parse: Got '{iteration_to_parse}'"
                             f"; but only '{n_iters}' iterations are available")
        index_condition = iteration_to_parse if iteration_to_parse < 0 else iteration_to_parse + 1  #1-based indexing in XPaths
    elif iteration_to_parse == 'all':
        if broken and n_iters >= 2:
            index_condition = {'<=': -2}
            if logger is not None:
                logger.info('The last parsed iteration is %s', n_iters - 2)
    else:
        if logger is not None:
            logger.error(
                "Invalid value for iteration_to_parse: Got '%s' "
                "Valid values are: 'first', 'last', 'all', or int", iteration_to_parse)
        raise ValueError(f"Invalid value for iteration_to_parse: Got '{iteration_to_parse}' "
                         "Valid values are: 'first', 'last', 'all', or int")

    filters = {}
    if index_condition is not None:
        filters['iteration'] = {'index': index_condition}
    return filters


MigrationDict: TypeAlias = "dict[str, dict[str, Literal['compatible'] | Callable]]"
"""
Type describing the dictionary defining the migration pathways
"""


def _find_migration(start: str, target: str, migrations: MigrationDict) -> list[Callable] | None:
    """
    Tries to find a migration path from the start to the target version
    via the defined migration functions

    :param start: str of the starting version
    :param target: str of the target version
    :param migrations: dict of funcs registered via the register_migration_function decorator

    :returns: list of migration functions to be called to go from start to target
    """

    if start == target:
        return []

    if start not in migrations:
        return None

    possible_migrations = migrations[start]
    if target in possible_migrations:
        migration = possible_migrations[target]
        if isinstance(migration, str) and migration == 'compatible':
            return []
        return [migration]

    for migrated_version, migration in possible_migrations.items():
        new_call_list = _find_migration(migrated_version, target, migrations)
        if new_call_list is None:
            #Cannot migrate to target from this version
            continue

        if isinstance(migration, str) and migration == 'compatible':
            call_list = []
        else:
            call_list = [migration]
        call_list += new_call_list
        return call_list
    return None


class _TaskParser:
    """
    Representation of all known parsing tasks for the out.xml file

    When set up it will initialize the known default tasks and check if they work
    for the given output version

    Accessing definition of task example

    .. code-block:: python

        from masci_tools.io.parsers.fleur.fleur_outxml_parser import _TaskParser

        p = _TaskParser('0.33')
        totE_definition = p.tasks['total_energy']
    """

    PARSE_FUNCTIONS = {
        'attrib', 'text', 'allAttribs', 'parentAttribs', 'singleValue', 'exists', 'attrib_exists', 'numberNodes'
    }
    ALL_ATTRIBS_FUNCTIONS = {'allAttribs', 'parentAttribs', 'singleValue'}

    CONTROL_KEYS = {'_general', '_modes', '_minimal', '_special', '_conversions', '_optional', '_minimum_version'}
    REQUIRED_KEYS = {'parse_type'}
    REQUIRED_KEYS_XML_GETTER = {'parse_type', 'name'}
    REQUIRED_KEYS_UTIL = {'parse_type', 'path_spec'}
    ALLOWED_KEYS = {'parse_type', 'path_spec', 'subdict', 'overwrite_last', 'force_list', 'kwargs'}
    ALLOWED_KEYS_ALLATTRIBS = {'parse_type', 'path_spec', 'subdict', 'base_value', 'overwrite', 'flat', 'kwargs'}
    ALLOWED_KEYS_XML_GETTER = {'parse_type', 'name', 'kwargs', 'result_names'}

    _version = '0.4.0'
    migrations: MigrationDict = {}
    conversion_functions: dict[str, Callable] = {}

    def __init__(self, version: str, validate_defaults: bool = False) -> None:
        """
        Initialize the default parse tasks
        Terminates if the version is not marked as working with the default tasks

        :param version: str of the wanted output version
        :param task_file: optional, file to override default_parse_tasks
        :param validate_defaults: bool, if True all tasks from the default tasks
                                  are added one by one and are checked for
                                  inconsistent keys
        """
        from . import default_parse_tasks as tasks

        self.iteration_tasks: list[str] = []
        self.general_tasks: list[str] = []
        self.version = convert_str_version_number(version)

        tasks_dict: dict[str, dict[str, Any]] = copy.deepcopy(tasks.TASKS_DEFINITION)  #type: ignore[arg-type]
        if validate_defaults:
            #Manually add each task to make sure that there are no typos/inconsitencies in the keys
            self.tasks = {}
            for task_name, task in tasks_dict.items():
                self.add_task(task_name, task)
        else:
            self.tasks = tasks_dict

        working: set[str] = tasks.__working_out_versions__
        #Look if the base version is compatible if not look for a migration
        if version not in working:

            working_version_tuples = {convert_str_version_number(v) for v in working}
            if all(working_version < self.version for working_version in working_version_tuples):
                warnings.warn(
                    f"Output version '{version}' is not explicitly stated as 'working'\n"
                    'with the current version of the outxml_parser.\n'
                    'Since the given version is newer than the latest working version\n'
                    'I will continue. Errors and warnings can occur!', UserWarning)
            else:
                base: str = tasks.__base_version__
                migration_list = _find_migration(base, version, self.migrations)

                if migration_list is None:
                    raise ValueError(f'Unsupported output version: {version}')

                for migration in migration_list:
                    self.tasks = migration(self.tasks)

    @property
    def optional_tasks(self) -> set[str]:
        """
        Return a set of the available optional defined tasks
        """
        return {key for key, val in self.tasks.items() if val.get('_optional', False)}

    def add_task(self,
                 task_name: str,
                 task_definition: dict[str, Any],
                 append: bool = False,
                 overwrite: bool = False) -> None:
        """
        Add a new task definition to the tasks dictionary

        Will first check if the definition has all the required keys

        :param task_name: str, key in the tasks dict
        :param task_definition: dict with the defined tasks
        :param overwrite: bool (optional), if True and the key is present in the dictionary it will be
                          overwritten with the new definition
        :param append: bool (optional), if True and the key is present in the dictionary the new definitions
                       will be inserted into this dictionary (inner keys WILL BE OVERWRITTEN). Additionally
                       if an inner key is overwritten with an empty dict the inner key will be removed

        The following keys are expected in each entry of the task_definition dictionary:
            :param parse_type: str, defines which methods to use when extracting the information
            :param path_spec: dict with all the arguments that should be passed to tag_xpath
                              or attrib_xpath to get the correct path
            :param subdict: str, if present the parsed values are put into this key in the output dictionary
            :param overwrite_last: bool, if True no list is inserted and each entry overwrites the last

        For the allAttribs parse_type there are more keys that can appear:
            :param base_value: str, optional. If given the attribute
                               with this name will be inserted into the key from the task_definition
                               all other keys are formatted as {task_key}_{attribute_name}
            :param ignore: list of str, these attributes will be ignored
            :param overwrite: list of str, these attributes will not create a list and overwrite any value
                              that might be there
            :param flat: bool, if False the dict parsed from the tag is inserted as a dict into the correspondin key
                               if True the values will be extracted and put into the output dictionary with the
                               format {task_key}_{attribute_name}

        """

        if task_name in self.tasks and not (append or overwrite):
            raise ValueError(f"Task '{task_name}' is already defined."
                             'Use append=True to append them (conflicting keys are overwritten)'
                             'or overwrite=True to remove all existing tasks')

        for task_key, definition in task_definition.items():

            if task_key.startswith('_'):
                if task_key not in self.CONTROL_KEYS:
                    raise ValueError(f'Unknown control key: {task_key}')
                continue

            task_keys = set(definition.keys())

            if not task_keys and task_key in self.tasks[task_name]:
                continue

            parse_type = definition['parse_type']
            if parse_type not in self.PARSE_FUNCTIONS | {'xmlGetter'}:
                raise ValueError(f'Unknown parse_type: {parse_type}')

            required = self.REQUIRED_KEYS
            allowed = self.ALLOWED_KEYS
            if parse_type == 'xmlGetter':
                required = required | self.REQUIRED_KEYS_XML_GETTER
                allowed = self.ALLOWED_KEYS_XML_GETTER
            else:
                required = required | self.REQUIRED_KEYS_UTIL
                if parse_type in self.ALL_ATTRIBS_FUNCTIONS:
                    allowed = self.ALLOWED_KEYS_ALLATTRIBS

            missing_required = required.difference(task_keys)
            if missing_required:
                raise ValueError(f'Reqired Keys missing: {missing_required}')

            extra_keys = task_keys.difference(allowed)
            if extra_keys:
                raise ValueError(f'Got extra Keys: {extra_keys}')

        if append:
            self.tasks.setdefault(task_name, {})
            for key, definition in task_definition.items():
                if definition:
                    self.tasks[task_name][key] = definition
                elif key in self.tasks[task_name]:
                    self.tasks[task_name].pop(key)
        else:
            self.tasks[task_name] = task_definition

    def determine_tasks(self,
                        fleurmodes: dict[str, Any],
                        optional_tasks: Iterable[str] | None = None,
                        minimal: bool = False,
                        iteration_to_parse: Literal['all', 'last', 'first'] | int = 'last') -> None:
        """
        Determine, which tasks to perform based on the fleur_modes

        :param fleurmodes: dict with the calculation modes
        :param minimal: bool, whether to only perform minimal tasks
        """

        if optional_tasks is None:
            optional_tasks = set()

        unknown = {name for name in optional_tasks if name not in self.optional_tasks}
        if unknown:
            raise ValueError(f"Unknown optional task(s): '{unknown}'\n"
                             f'The following are available: {self.optional_tasks}')

        for task_name, definition in self.tasks.items():

            if '_minimum_version' in definition:
                min_version = convert_str_version_number(definition['_minimum_version'])
                if self.version < min_version:
                    continue

            optional = definition.get('_optional', False)
            if optional and task_name not in optional_tasks:
                continue

            if minimal and not definition.get('_minimal', False):
                continue

            #These tasks are always added manually
            if definition.get('_special', False):
                continue

            requirements = definition.get('_modes', [])
            check = [fleurmodes[mode] == required_value for mode, required_value in requirements]
            if not all(check):
                continue

            if definition.get('_general', False):
                self.general_tasks.append(task_name)
            else:
                self.iteration_tasks.append(task_name)

        #Manual overrides for certain fleur modes
        if fleurmodes['dos'] or fleurmodes['band'] or fleurmodes['cf_coeff']:
            self.iteration_tasks = ['iteration_number', 'fermi_energy']
            if fleurmodes['bz_integration'] == 'hist':
                self.iteration_tasks = ['iteration_number', 'fermi_energy', 'bandgap']

        if fleurmodes['plot']:
            self.iteration_tasks = []  #In this case there are multiple possibilities where fleur terminates
            #So we discard all the iteration tasks

        if fleurmodes['relax'] and iteration_to_parse == 'last':
            if 'distances' in self.iteration_tasks:
                self.iteration_tasks.remove('distances')
            if 'magnetic_distances' in self.iteration_tasks:
                self.iteration_tasks.remove('magnetic_distances')
            if 'nmmp_distances' in self.iteration_tasks:
                self.iteration_tasks.remove('nmmp_distances')

    def perform_task(self, task_name: str, context: _EvalContext, out_dict: dict, use_lists: bool = True) -> dict:
        """
        Evaluates the task given in the tasks_definition dict

        :param task_name: str, specifies the task to perform
        :param node: etree.Element, the xpath expressions are evaluated from this node
        :param out_dict: dict, output will be put in this dictionary
        :param schema_dict: dict, here all paths and attributes are stored according to the
                            outputschema
        :param constants: dict with all the defined mathematical constants
        :param logger: logger object for logging warnings, errors
        :param root_tag: str, this string will be appended in front of any xpath before it is evaluated
        :param use_lists: bool, if True lists are created for each key if not otherwise specified

        """
        from masci_tools.io.common_functions import camel_to_snake

        #TODO: Could be moved into _EvalContext.__getitem__
        parse_functions = {
            'attrib': context.attribute,
            'text': context.text,
            'allAttribs': context.all_attributes,
            'parentAttribs': context.parent_attributes,
            'singleValue': context.single_value,
            'exists': context.tag_exists,
            'attrib_exists': context.attribute_exists,
            'numberNodes': context.number_nodes
        }

        try:
            tasks_definition = self.tasks[task_name]
        except KeyError as exc:
            raise KeyError(f'Unknown Task: {task_name}') from exc

        for task_key, spec in tasks_definition.items():

            if task_key.startswith('_'):
                continue

            if spec['parse_type'] == 'xmlGetter':
                action = getattr(xml_getters, spec['name'])
                action = partial(action,
                                 context.node,
                                 context.schema_dict,
                                 logger=context.logger,
                                 constants=context.constants)
                args = spec.get('kwargs', {}).copy()
            else:
                action = parse_functions[spec['parse_type']]
                args = spec['path_spec'].copy()
                args = {**args, **spec.get('kwargs', {})}

            if spec['parse_type'] == 'singleValue':
                args.setdefault('ignore', []).append('comment')

            parsed_dict = out_dict
            if 'subdict' in spec:
                parsed_dict = out_dict.setdefault(spec['subdict'], {})

            parsed_value = action(**args)

            if spec['parse_type'] == 'xmlGetter' and 'result_names' in spec:
                if isinstance(parsed_value, tuple):
                    if len(spec['result_names']) != len(parsed_value):
                        raise ValueError('Wrong number of result names given.'
                                         f"Got {len(parsed_value)} values and {len(spec['result_names'])} names")
                    parsed_value = dict(zip(spec['result_names'], parsed_value))
                else:
                    task_key = spec['result_names'][0]

            if isinstance(parsed_value, dict):

                if spec['parse_type'] == 'singleValue':
                    base_value = 'value'
                    no_list = ['units']
                    flat = True
                elif spec['parse_type'] in ['allAttribs', 'parentAttribs']:
                    base_value = spec.get('base_value', '')
                    no_list = spec.get('overwrite', [])
                    flat = spec.get('flat', True)

                if flat:
                    for key, val in parsed_value.items():
                        current_key = f'{task_key}_{camel_to_snake(key)}' if key != base_value else task_key
                        if key in no_list or not use_lists:
                            parsed_dict[current_key] = val
                        else:
                            parsed_dict.setdefault(current_key, []).append(val)

                else:
                    parsed_dict[task_key] = {camel_to_snake(key): val for key, val in parsed_value.items()}

            else:
                overwrite = spec.get('overwrite_last', False)
                force_list = spec.get('force_list', use_lists)
                if force_list and not overwrite:
                    parsed_dict.setdefault(task_key, []).append(parsed_value)
                else:
                    parsed_dict[task_key] = parsed_value if parsed_value is not None else parsed_dict.get(task_key)

        conversions = tasks_definition.get('_conversions', [])
        for conversion in conversions:
            if not isinstance(conversion, Conversion):
                warnings.warn(
                    'Providing the _conversions as a list of strings is deprecated'
                    'Use the Conversion namedtuple from masci_tools.util.parse_utils instead', DeprecationWarning)
                conversion = Conversion(name=conversion)

            action = self.conversion_functions[conversion.name]
            out_dict = action(out_dict, *conversion.args, logger=context.logger, **conversion.kwargs)

        return out_dict


F = TypeVar('F', bound=Callable[..., Any])
"""Generic Callable type"""


def register_migration(base_version: str, target_version: str | list[str]) -> Callable[[F], F]:
    """
    Decorator to add migration for task definition dictionary to the _TaskParser class
    The function should only take the dict of task definitions as an argument

    :param base_version: str of the version, from which the migration starts
    :param target_version: str or list of str with the versions that work after
                           the migration has been performed

    """

    def migration_decorator(func: F) -> F:
        """
        Return decorated _TaskParser object with migrations dict attribute
        Here all registered migrations are inserted
        """

        target_version_list = target_version
        if not isinstance(target_version_list, list):
            target_version_list = [target_version_list]

        for valid_version in target_version_list:
            _TaskParser.migrations.setdefault(base_version, {})[valid_version] = func

            for valid_version_2 in target_version_list:
                if valid_version == valid_version_2:
                    continue
                if int(valid_version.split('.')[1]) > int(valid_version_2.split('.')[1]):
                    _TaskParser.migrations.setdefault(valid_version, {})[valid_version_2] = 'compatible'
                else:
                    _TaskParser.migrations.setdefault(valid_version_2, {})[valid_version] = 'compatible'

        return func

    return migration_decorator


def conversion_function(func: F) -> F:
    """
    Marks a function as a conversion function, which can be called after
    performing a parsing task. The function can be specified via the _conversions
    control key in the task definitions.

    A conversion function has to have the following arguments:
        :param out_dict: dict with the previously parsed information
        :param parser_info_out: dict, with warnings, info, errors, ...

    and return only the modified output dict
    """
    _TaskParser.conversion_functions[func.__name__] = func

    return func



"""
Main module defining the CLI for parts of the masci-tools repository
"""
import click
import click_completion
from masci_tools import __version__

click_completion.init()

# Activate the completion of parameter types provided by the click_completion package
# for bash: eval "$(_MASCI_TOOLS_COMPLETE=source masci-tools)"


@click.group('masci-tools', context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(__version__, '-v', '--version', message='masci-tools version %(version)s')
def cli():
    """CLI for the `masci-tools` library."""



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains a class for organizing and grouping changes to a input file
of fleur in a robust way.

Essentially a low-level version of the FleurinpModifier in aiida_fleur.
"""
from __future__ import annotations

from typing import Any, Callable, NamedTuple
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  #type: ignore

from masci_tools.util.xml.collect_xml_setters import XPATH_SETTERS, SCHEMA_DICT_SETTERS, NMMPMAT_SETTERS
from masci_tools.util.xml.common_functions import clear_xml
from masci_tools.io.fleur_xml import load_inpxml
from masci_tools.util.typing import XMLFileLike, FileLike
from pathlib import Path
from lxml import etree
import warnings
#Enable warnings for missing docstrings
#pylint: enable=missing-function-docstring


class ModifierTask(NamedTuple):
    name: str
    args: tuple[Any, ...] = ()
    kwargs: dict[str, Any] = {}


class FleurXMLModifier:
    """
    Class for grouping and organizing changes to a inp.xml file of fleur via the
    xml setting methods in :py:mod:`~masci_tools.util.xml.xml_setters_names` and
    :py:mod:`~masci_tools.util.xml.xml_setters_basic`

    The basic usage is shown below

    .. code-block:: python

        from masci_tools.io.fleurxmlmodifier import FleurXMLModifier

        fmode = FleurXMLModifier()

        #Add changes by calling the methods on this class
        #(names correspond to the setting methods in the xml_setters modules)
        #They are not modifying a input file directly
        #Instead all the tasks are collected and performed in one go

        fmode.set_inpchanges({'Kmax': 4.0}) #Set Kmax to 4.0
        fmode.shift_value({'Gmax': 5.0}) #Add 5 to the current value of Gmax

        #Set the local orbital configuration on all iron atoms to '3s 3p'
        fmode.set_species('all-Fe', {'lo': [{'n':3, 'l': 's', 'type': 'SCLO'},
                                            {'n':3, 'l': 'p', 'type': 'SCLO'}]})

        #To undo the last change call undo
        #fmode.undo()

        #revert_all=True resets all added tasks
        #fmode.undo(revert_all=True)

        #To apply the changes to an input file use the modify_xmlfile method
        new_xmltree = fmode.modify_xmlfile('/path/to/input/file/inp.xml')

    """

    _xpath_functions: dict[str, Callable] = XPATH_SETTERS
    _schema_dict_functions: dict[str, Callable] = SCHEMA_DICT_SETTERS
    _nmmpmat_functions: dict[str, Callable] = NMMPMAT_SETTERS

    _extra_functions: dict[str, dict[Literal['xpath', 'schema_dict', 'nmmpmat'], Callable]] = {}

    def __new__(cls, validate_signatures=True):

        if getattr(cls, 'xpath_functions', None) is None:
            cls.xpath_functions = {**cls._xpath_functions, **cls._extra_functions.get('xpath', {})}
            cls.schema_dict_functions = {**cls._schema_dict_functions, **cls._extra_functions.get('schema_dict', {})}
            cls.nmmpmat_functions = {**cls._nmmpmat_functions, **cls._extra_functions.get('nmmpmat', {})}

        return super().__new__(cls)

    def __init__(self, validate_signatures: bool = True) -> None:

        self._tasks: list[ModifierTask] = []
        self.validate_signatures = validate_signatures

    @classmethod
    def fromList(cls, task_list: list[tuple[str, dict[str, Any]]], *args: Any, **kwargs: Any) -> FleurXMLModifier:
        """
        Instantiate the FleurXMLModifier from a list of tasks to be added immediately

        :param task_list: list of tuples first index is the name of the method
                          second is defining the arguments by keyword in a dict

        Other arguments are passed on to the __init__ method

        :returns: class with the task list instantiated
        """

        fm = cls(*args, **kwargs)
        fm.add_task_list(task_list)
        return fm

    def add_task_list(self, task_list: list[tuple[str, dict[str, Any]]]) -> None:
        """
        Add a list of tasks to be added

        :param task_list: list of tuples first index is the name of the method
                          second is defining the arguments by keyword in a dict
        """

        facade_methods = self.get_avail_actions()

        for name, kwargs in task_list:
            try:
                facade_methods[name](**kwargs)
            except KeyError as exc:
                raise ValueError(f"Unknown modification method '{name}'") from exc

    def _validate_signature(self, name: str, *args: Any, **kwargs: Any) -> None:
        """
        DEPRECATED: use `_validate_arguments` instead without unpacking args/kwargs
        """
        warnings.warn(
            'The _validate_signature method is deprecated. '
            'Please use _validate_arguments without unpacking args/kwargs instead', DeprecationWarning)
        self._validate_arguments(name, args, kwargs)

    def _get_setter_function_and_prefix(self, name: str) -> tuple[Callable[[Any], Any], tuple[str, ...]]:
        """
        Get the setter function and a prefix standing in for the arguments that
        are substituted when performing the modification
        """
        if name in self.xpath_functions:
            func = self.xpath_functions[name]
            prefix: tuple[str, ...] = ('xmltree',)
        elif name in self.schema_dict_functions:
            func = self.schema_dict_functions[name]
            prefix = ('xmltree', 'schema_dict')
        elif name in self.nmmpmat_functions:
            func = self.nmmpmat_functions[name]
            prefix = ('xmltree', 'nmmplines', 'schema_dict')

        if func is None:
            raise ValueError(f'Failed to validate setter {name}. Maybe the function was'
                             'not registered in masci_tools.util.xml.collect_xml_setters')

        #For functions decorated with the schema_dict_version_dispatch
        #We check only the default (This function should have a compatible signature for all registered functions)
        if getattr(func, 'registry', None) is not None:
            func = func.registry['default']

        return func, prefix

    def _get_setter_func_kwargs(self, name: str, args: tuple[Any, ...], kwargs: dict[str, Any]) -> dict[str, Any]:
        """
        Map the given args and kwargs to just kwargs for the
        setter function with the given name

        :param name: name of the setter function
        :param args: positional arguments to the setter function
        :param kwargs: keyword arguments to the setter function
        """
        from inspect import signature
        func, prefix = self._get_setter_function_and_prefix(name)

        sig = signature(func)
        bound = sig.bind(*prefix, *args, **kwargs)

        return {k: v for k, v in bound.arguments.items() if k not in ('xmltree', 'nmmplines', 'schema_dict')}

    def _validate_arguments(self, name: str, args: tuple[Any, ...], kwargs: dict[str, Any]) -> None:
        """
        Validate that the given arguments to the registration
        method can be used to call the corresponding XML modifying function
        """
        from inspect import signature

        if self.validate_signatures:
            func, prefix = self._get_setter_function_and_prefix(name)
            try:
                sig = signature(func)
                sig.bind(*prefix, *args, **kwargs)
            except TypeError as exc:
                raise TypeError(
                    f"The given arguments for the registration method '{name}' are not valid for the XML modifying function"
                    f'The following error was raised: {exc}') from exc

    @classmethod
    def apply_modifications(cls,
                            xmltree: etree._ElementTree,
                            nmmp_lines: list[str] | None,
                            modification_tasks: list[ModifierTask],
                            validate_changes: bool = True) -> tuple[etree._ElementTree, list[str] | None]:
        """
        Applies given modifications to the fleurinp lxml tree.
        It also checks if a new lxml tree is validated against schema.
        Does not rise an error if inp.xml is not validated, simple prints a message about it.

        :param xmltree: a lxml tree to be modified (IS MODIFIED INPLACE)
        :param nmmp_lines: a n_mmp_mat file to be modified (IS MODIFIED INPLACE)
        :param modification_tasks: a list of modification tuples
        :param validate_changes: bool optional (default True), if True after all tasks are performed
                                 both the xmltree and nmmp_lines are checked for consistency

        :returns: a modified lxml tree and a modified n_mmp_mat file
        """
        from masci_tools.util.xml.xml_setters_nmmpmat import validate_nmmpmat

        xmltree, schema_dict = load_inpxml(xmltree)
        xmltree, _ = clear_xml(xmltree)

        for task in modification_tasks:
            if task.name in cls.xpath_functions:
                action = cls.xpath_functions[task.name]
                xmltree = action(xmltree, *task.args, **task.kwargs)

            elif task.name in cls.schema_dict_functions:
                action = cls.schema_dict_functions[task.name]
                xmltree = action(xmltree, schema_dict, *task.args, **task.kwargs)

            elif task.name in cls.nmmpmat_functions:
                action = cls.nmmpmat_functions[task.name]
                nmmp_lines = action(xmltree, nmmp_lines, schema_dict, *task.args, **task.kwargs)

            else:
                raise ValueError(f'Unknown task {task.name}')

        if validate_changes:
            schema_dict.validate(xmltree, header='Changes were not valid')
            try:
                validate_nmmpmat(xmltree, nmmp_lines, schema_dict)
            except ValueError as exc:
                msg = f'Changes were not valid (n_mmp_mat file is not compatible): {modification_tasks}'
                raise ValueError(msg) from exc

        return xmltree, nmmp_lines

    @property
    def task_list(self) -> list[tuple[str, dict[str, Any]]]:
        """
        Return the current changes in a format accepted by :py:meth:`add_task_list()`
        and :py:meth:`fromList()`
        """

        tasks = []
        for change in self._tasks:
            #Here we already validated the arguments so we know we can just get the kwargs
            kwargs = self._get_setter_func_kwargs(change.name, change.args, change.kwargs)
            tasks.append((change.name, kwargs))

        return tasks

    def get_avail_actions(self) -> dict[str, Callable]:
        """
        Returns the allowed functions from FleurXMLModifier
        """
        outside_actions = {
            'set_inpchanges': self.set_inpchanges,
            'shift_value': self.shift_value,
            'set_species': self.set_species,
            'set_species_label': self.set_species_label,
            'clone_species': self.clone_species,
            'switch_species': self.switch_species,
            'switch_species_label': self.switch_species_label,
            'shift_value_species_label': self.shift_value_species_label,
            'set_atomgroup': self.set_atomgroup,
            'set_atomgroup_label': self.set_atomgroup_label,
            'set_complex_tag': self.set_complex_tag,
            'set_simple_tag': self.set_simple_tag,
            'create_tag': self.create_tag,
            'delete_tag': self.delete_tag,
            'delete_att': self.delete_att,
            'replace_tag': self.replace_tag,
            'set_text': self.set_text,
            'set_first_text': self.set_first_text,
            'set_attrib_value': self.set_attrib_value,
            'set_first_attrib_value': self.set_first_attrib_value,
            'add_number_to_attrib': self.add_number_to_attrib,
            'add_number_to_first_attrib': self.add_number_to_first_attrib,
            'xml_create_tag': self.xml_create_tag,
            'xml_replace_tag': self.xml_replace_tag,
            'xml_delete_tag': self.xml_delete_tag,
            'xml_delete_att': self.xml_delete_att,
            'xml_set_attrib_value_no_create': self.xml_set_attrib_value_no_create,
            'xml_set_text_no_create': self.xml_set_text_no_create,
            'set_nmmpmat': self.set_nmmpmat,
            'rotate_nmmpmat': self.rotate_nmmpmat,
            'align_nmmpmat_to_sqa': self.align_nmmpmat_to_sqa,
            'set_nkpts': self.set_nkpts,
            'set_kpath': self.set_kpath,
            'set_kpointlist': self.set_kpointlist,
            'switch_kpointset': self.switch_kpointset,
        }
        return outside_actions

    def undo(self, revert_all: bool = False) -> list[ModifierTask]:
        """
        Cancels the last change or all of them

        :param revert_all: set True if need to cancel all the changes, False if the last one.
        """
        if revert_all:
            self._tasks = []
        else:
            if self._tasks:
                self._tasks.pop()
                #TODO delete nodes from other nodes
                #del self._tasks[-1]
        return self._tasks

    def changes(self) -> list[ModifierTask]:
        """
        Prints out all changes currently registered on this instance
        """
        from pprint import pprint
        pprint(self._tasks)
        return self._tasks

    def modify_xmlfile(self,
                       original_inpxmlfile: XMLFileLike,
                       original_nmmp_file: FileLike | list[str] | None = None,
                       validate_changes: bool = True) -> tuple[etree._ElementTree, list[str]] | etree._ElementTree:
        """
        Applies the registered modifications to a given inputfile

        :param original_inpxmlfile: either path to the inp.xml file, opened file handle
                                    or a xml etree to be parsed
        :param original_nmmp_file: path or list of str to a corresponding density matrix
                                   file

        :raises ValueError: if the parsing of the input file

        :returns: a modified xmltree and if existent a modified density matrix file
        """
        original_xmltree, _ = load_inpxml(original_inpxmlfile)

        if original_nmmp_file is not None:
            if isinstance(original_nmmp_file, (str, Path)):
                with open(original_nmmp_file, encoding='utf-8') as n_mmp_file:
                    original_nmmp_lines = n_mmp_file.read().split('\n')
            else:
                original_nmmp_lines = original_nmmp_file  #type:ignore[assignment]
        else:
            original_nmmp_lines = None

        new_xmltree, new_nmmp_lines = self.apply_modifications(original_xmltree,
                                                               original_nmmp_lines,
                                                               self._tasks,
                                                               validate_changes=validate_changes)

        if new_nmmp_lines is None:
            return new_xmltree
        return new_xmltree, new_nmmp_lines

    def set_inpchanges(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_inpchanges()` to
        the list of tasks that will be done on the xmltree.

        :param changes: a dictionary with changes
        :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute

        An example of changes::

            changes = {'itmax' : 1,
                       'l_noco': True,
                       'ctail': False,
                       'l_ss': True}
        """
        if 'change_dict' in kwargs:
            warnings.warn('The argument change_dict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('change_dict')
        self._validate_arguments('set_inpchanges', args, kwargs)
        self._tasks.append(ModifierTask('set_inpchanges', args, kwargs))

    def shift_value(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.shift_value()` to
        the list of tasks that will be done on the xmltree.

        :param changes: a python dictionary with the keys to shift and the shift values.
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                     `rel`/`relative` multiplies the old value with the given value
                     `abs`/`absolute` adds the old value and the given value
        :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute

        An example of changes::

                changes = {'itmax' : 1, 'dVac': -0.123}
        """
        if 'change_dict' in kwargs:
            warnings.warn('The argument change_dict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('change_dict')
        self._validate_arguments('shift_value', args, kwargs)
        self._tasks.append(ModifierTask('shift_value', args, kwargs))

    def set_species(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_species()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the specie you want to change
                             Can be name of the species, 'all' or 'all-<string>' (sets species with the string in the species name)
        :param changes: a python dict specifying what you want to change.
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        **changes** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a MT radius it
        can be done via::

            changes = {'mtSphere' : {'radius' : 2.2}}

        Another example::

            'changes': {'special': {'socscale': 0.0}}

        that switches SOC terms on a sertain specie. ``mtSphere``, ``atomicCutoffs``,
        ``energyParameters``, ``lo``, ``electronConfig``, ``nocoParams``, ``ldaU`` and
        ``special`` keys are supported. To find possible
        keys of the inner dictionary please refer to the FLEUR documentation flapw.de
        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        self._validate_arguments('set_species', args, kwargs)
        self._tasks.append(ModifierTask('set_species', args, kwargs))

    def set_species_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_species_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
        :param changes: a python dict specifying what you want to change.

        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        self._validate_arguments('set_species_label', args, kwargs)
        self._tasks.append(ModifierTask('set_species_label', args, kwargs))

    def clone_species(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.clone_species()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the specie you want to clone
                            Has to correspond to one single species (no 'all'/'all-<search_string>')
        :param new_name: new name of the cloned species
        :param changes: a optional python dict specifying what you want to change.
        """
        self._validate_arguments('clone_species', args, kwargs)
        self._tasks.append(ModifierTask('clone_species', args, kwargs))

    def switch_species(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.switch_species()` to
        the list of tasks that will be done on the xmltree.

        :param new_species_name: name of the species to switch to
        :param position: position of an atom group to be changed. If equals to 'all', all species will be changed
        :param species: atom groups, corresponding to the given species will be changed
        :param clone: if True and the new species name does not exist and it corresponds to changing
                  from one species the species will be cloned with :py:func:`clone_species()`
        :param changes: changes to do if the species is cloned
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details`
        """
        self._validate_arguments('switch_species', args, kwargs)
        self._tasks.append(ModifierTask('switch_species', args, kwargs))

    def switch_species_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.switch_species_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which group will be changed. 'all' to change all the groups
        :param new_species_name: name of the species to switch to
        :param clone: if True and the new species name does not exist and it corresponds to changing
                  from one species the species will be cloned with :py:func:`clone_species()`
        :param changes: changes to do if the species is cloned
        """
        self._validate_arguments('switch_species_label', args, kwargs)
        self._tasks.append(ModifierTask('switch_species_label', args, kwargs))

    def shift_value_species_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.shift_value_species_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' if set up all species
        :param attribute_name: name of the attribute to change
        :param number_to_add: value to add or to multiply by
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                    `rel`/`relative` multiplies the old value with `number_to_add`
                    `abs`/`absolute` adds the old value and `number_to_add`

        Kwargs if the attribute_name does not correspond to a unique path:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use attribute_name instead', DeprecationWarning)
            kwargs['attribute_name'] = kwargs.pop('attributename')
        if 'value_given' in kwargs:
            warnings.warn('The argument value_given is deprecated. Use number_to_add instead', DeprecationWarning)
            kwargs['number_to_add'] = kwargs.pop('value_given')
        self._validate_arguments('shift_value_species_label', args, kwargs)
        self._tasks.append(ModifierTask('shift_value_species_label', args, kwargs))

    def set_atomgroup(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup()` to
        the list of tasks that will be done on the xmltree.

        :param changes: a python dict specifying what you want to change.
        :param position: position of an atom group to be changed. If equals to 'all', all species will be changed
        :param species: atom groups, corresponding to the given species will be changed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

        **changes** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
        can be done via::

            'changes': {'nocoParams': {'beta': val}}

        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        if 'create' in kwargs:
            warnings.warn('The argument create is deprecated and is ignored.', DeprecationWarning)
            kwargs.pop('create')
        self._validate_arguments('set_atomgroup', args, kwargs)
        self._tasks.append(ModifierTask('set_atomgroup', args, kwargs))

    def set_atomgroup_label(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
        :param changes: a python dict specifying what you want to change.

        **changes** is a python dictionary containing dictionaries that specify attributes
        to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
        can be done via::

            'changes': {'nocoParams': {'beta': val}}

        """
        if 'attributedict' in kwargs:
            warnings.warn('The argument attributedict is deprecated. Use changes instead', DeprecationWarning)
            kwargs['changes'] = kwargs.pop('attributedict')
        if 'create' in kwargs:
            warnings.warn('The argument create is deprecatedand is ignored.', DeprecationWarning)
            kwargs.pop('create')
        self._validate_arguments('set_atomgroup_label', args, kwargs)
        self._tasks.append(ModifierTask('set_atomgroup_label', args, kwargs))

    def create_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.create_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag: str of the tag to create or etree Element with the same name
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create_parents: bool optional (default False), if True and the given xpath has no results the
                               the parent tags are created recursively
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('create_tag', args, kwargs)
        self._tasks.append(ModifierTask('create_tag', args, kwargs))

    def delete_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.delete_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag: str of the tag to delete
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurrence of the parent nodes to delete a tag.
                            By default all nodes are used.
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('delete_tag', args, kwargs)
        self._tasks.append(ModifierTask('delete_tag', args, kwargs))

    def delete_att(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.delete_att()` to
        the list of tasks that will be done on the xmltree.

        :param name: str of the attribute to delete
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurrence of the parent nodes to delete a attribute.
                            By default all nodes are used.
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
        """
        if 'attrib_name' in kwargs:
            warnings.warn('The argument attrib_name is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attrib_name')
        self._validate_arguments('delete_att', args, kwargs)
        self._tasks.append(ModifierTask('delete_att', args, kwargs))

    def replace_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.replace_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag: str of the tag to replace
        :param element: a new tag
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurrence of the parent nodes to replace a tag.
                            By default all nodes are used.
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        if 'newelement' in kwargs:
            warnings.warn('The argument newelement is deprecated. Use element instead', DeprecationWarning)
            kwargs['element'] = kwargs.pop('newelement')
        self._validate_arguments('replace_tag', args, kwargs)
        self._tasks.append(ModifierTask('replace_tag', args, kwargs))

    def set_complex_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_complex_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: name of the tag to set
        :param changes: Keys in the dictionary correspond to names of tags and the values are the modifications
                        to do on this tag (attributename, subdict with changes to the subtag, ...)
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create: bool optional (default False), if True and the path, where the complex tag is
                       set does not exist it is created
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        self._validate_arguments('set_complex_tag', args, kwargs)
        self._tasks.append(ModifierTask('set_complex_tag', args, kwargs))

    def set_simple_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_simple_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: str name of the tag to modify/set
        :param changes: list of dicts or dict with the changes. Elements in list describe multiple tags.
                        Keys in the dictionary correspond to {'attributename': attributevalue}
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create_parents: bool optional (default False), if True and the path, where the simple tags are
                               set does not exist it is created
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """
        self._validate_arguments('set_simple_tag', args, kwargs)
        self._tasks.append(ModifierTask('set_simple_tag', args, kwargs))

    def set_text(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_text()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: str name of the tag, where the text should be set
        :param text: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.
        :param create: bool optional (default False), if True the tag is created if is missing
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        self._validate_arguments('set_text', args, kwargs)
        self._tasks.append(ModifierTask('set_text', args, kwargs))

    def set_first_text(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_first_text()` to
        the list of tasks that will be done on the xmltree.

        :param tag_name: str name of the tag, where the text should be set
        :param text: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create: bool optional (default False), if True the tag is created if is missing
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        self._validate_arguments('set_first_text', args, kwargs)
        self._tasks.append(ModifierTask('set_first_text', args, kwargs))

    def set_attrib_value(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_attrib_value()` to
        the list of tasks that will be done on the xmltree.

        :param name: the attribute name to set
        :param value: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.
        :param create: bool optional (default False), if True the tag is created if is missing
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'attribv' in kwargs:
            warnings.warn('The argument attribv is deprecated. Use value instead', DeprecationWarning)
            kwargs['value'] = kwargs.pop('attribv')
        self._validate_arguments('set_attrib_value', args, kwargs)
        self._tasks.append(ModifierTask('set_attrib_value', args, kwargs))

    def set_first_attrib_value(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_first_attrib_value()` to
        the list of tasks that will be done on the xmltree.

        :param name: the attribute name to set
        :param value: value or list of values to set
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create: bool optional (default False), if True the tag is created if is missing
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'attribv' in kwargs:
            warnings.warn('The argument attribv is deprecated. Use value instead', DeprecationWarning)
            kwargs['value'] = kwargs.pop('attribv')
        self._validate_arguments('set_first_attrib_value', args, kwargs)
        self._tasks.append(ModifierTask('set_first_attrib_value', args, kwargs))

    def add_number_to_attrib(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.add_number_to_attrib()` to
        the list of tasks that will be done on the xmltree.

        :param name: the attribute name to change
        :param number_to_add: number to add/multiply with the old attribute value
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                    `rel`/`relative` multiplies the old value with `number_to_add`
                    `abs`/`absolute` adds the old value and `number_to_add`
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'add_number' in kwargs:
            warnings.warn('The argument add_number is deprecated. Use number_to_add instead', DeprecationWarning)
            kwargs['number_to_add'] = kwargs.pop('add_number')
        self._validate_arguments('add_number_to_attrib', args, kwargs)
        self._tasks.append(ModifierTask('add_number_to_attrib', args, kwargs))

    def add_number_to_first_attrib(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.add_number_to_first_attrib()` to
        the list of tasks that will be done on the xmltree.

        :param name: the attribute name to change
        :param number_to_add: number to add/multiply with the old attribute value
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param mode: str (either `rel`/`relative` or `abs`/`absolute`).
                    `rel`/`relative` multiplies the old value with `number_to_add`
                    `abs`/`absolute` adds the old value and `number_to_add`
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details


        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other

        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'add_number' in kwargs:
            warnings.warn('The argument add_number is deprecated. Use number_to_add instead', DeprecationWarning)
            kwargs['number_to_add'] = kwargs.pop('add_number')
        self._validate_arguments('add_number_to_first_attrib', args, kwargs)
        self._tasks.append(ModifierTask('add_number_to_first_attrib', args, kwargs))

    def xml_create_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_create_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path where to place a new tag
        :param element: a tag name or etree Element to be created
        :param place_index: defines the place where to put a created tag
        :param tag_order: defines a tag order
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        self._validate_arguments('xml_create_tag', args, kwargs)
        self._tasks.append(ModifierTask('xml_create_tag', args, kwargs))

    def xml_replace_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_replace_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path to the tag to be replaced
        :param element: a new tag
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        if 'newelement' in kwargs:
            warnings.warn('The argument newelement is deprecated. Use element instead', DeprecationWarning)
            kwargs['element'] = kwargs.pop('newelement')

        self._validate_arguments('xml_replace_tag', args, kwargs)
        self._tasks.append(ModifierTask('xml_replace_tag', args, kwargs))

    def xml_delete_tag(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_delete_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path to the tag to be deleted
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        self._validate_arguments('xml_delete_tag', args, kwargs)
        self._tasks.append(ModifierTask('xml_delete_tag', args, kwargs))

    def xml_delete_att(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_delete_att()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path to the attribute to be deleted
        :param name: the name of an attribute
        :param occurrences: int or list of int. Which occurrence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        self._validate_arguments('xml_delete_att', args, kwargs)
        self._tasks.append(ModifierTask('xml_delete_att', args, kwargs))

    def xml_set_attrib_value_no_create(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_set_attrib_value_no_create()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path where to set the attributes
        :param name: the attribute name to set
        :param value: value or list of values to set (if not str they will be converted with `str(value)`)
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.
        """
        if 'attributename' in kwargs:
            warnings.warn('The argument attributename is deprecated. Use name instead', DeprecationWarning)
            kwargs['name'] = kwargs.pop('attributename')
        if 'attribv' in kwargs:
            warnings.warn('The argument attribv is deprecated. Use value instead', DeprecationWarning)
            kwargs['value'] = kwargs.pop('attribv')
        self._validate_arguments('xml_set_attrib_value_no_create', args, kwargs)
        self._tasks.append(ModifierTask('xml_set_attrib_value_no_create', args, kwargs))

    def xml_set_text_no_create(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_set_text_no_create()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path where to set the attributes
        :param text: value or list of values to set (if not str they will be converted with `str(value)`)
        :param occurrences: int or list of int. Which occurrence of the node to set. By default all are set.
        """
        self._validate_arguments('xml_set_text_no_create', args, kwargs)
        self._tasks.append(ModifierTask('xml_set_text_no_create', args, kwargs))

    def set_nmmpmat(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.set_nmmpmat()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the species you want to change
        :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
        :param spin: integer, specifies which spin block should be modified
        :param state_occupations: list, sets the diagonal elements of the density matrix and everything
                          else to zero
        :param denmat: matrix, specify the density matrix explicitly
        :param phi: float, optional angle (radian), by which to rotate the density matrix before writing it
        :param theta: float, optional angle (radian), by which to rotate the density matrix before writing it
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        """
        self._validate_arguments('set_nmmpmat', args, kwargs)
        self._tasks.append(ModifierTask('set_nmmpmat', args, kwargs))

    def rotate_nmmpmat(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.rotate_nmmpmat()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the species you want to change
        :param orbital: integer or string ('all'), orbital quantum number of the LDA+U procedure to be modified
        :param phi: float, angle (radian), by which to rotate the density matrix
        :param theta: float, angle (radian), by which to rotate the density matrix
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        """
        self._validate_arguments('rotate_nmmpmat', args, kwargs)
        self._tasks.append(ModifierTask('rotate_nmmpmat', args, kwargs))

    def align_nmmpmat_to_sqa(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.align_nmmpmat_to_sqa()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the species you want to change
        :param orbital: integer or string ('all'), orbital quantum number of the LDA+U procedure to be modified
        :param phi_before: float or list of floats, angle (radian),
                           values for phi for the previous alignment of the density matrix
        :param theta_before: float or list of floats, angle (radian),
                             values for theta for the previous alignment of the density matrix
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
        """
        self._validate_arguments('align_nmmpmat_to_sqa', args, kwargs)
        self._tasks.append(ModifierTask('align_nmmpmat_to_sqa', args, kwargs))

    def set_kpointlist(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_kpointlist()` to
        the list of tasks that will be done on the xmltree.

        .. warning::
            For input versions Max4 and older **all** keyword arguments are not valid (`name`, `kpoint_type`,
            `special_labels`, `switch` and `overwrite`)

        :param kpoints: list or array containing the **relative** coordinates of the kpoints
        :param weights: list or array containing the weights of the kpoints
        :param name: str for the name of the list, if not given a default name is generated
        :param kpoint_type: str specifying the type of the kPointList ('path', 'mesh', 'spex', 'tria', ...)
        :param special_labels: dict mapping indices to labels. The labels will be inserted for the kpoints
                               corresponding to the given index
        :param switch: bool, if True the kPointlist will be used by Fleur when starting the next calculation
        :param overwrite: bool, if True and a kPointlist with the given name already exists it will be overwritten
        """
        self._validate_arguments('set_kpointlist', args, kwargs)
        self._tasks.append(ModifierTask('set_kpointlist', args, kwargs))

    def switch_kpointset(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.switch_kpointset()` to
        the list of tasks that will be done on the xmltree.

        .. warning::
            This method is only supported for input versions after the Max5 release

        :param list_name: name of the kPoint set to use
        """
        self._validate_arguments('switch_kpointset', args, kwargs)
        self._tasks.append(ModifierTask('switch_kpointset', args, kwargs))

    def set_nkpts(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_nkpts()` to
        the list of tasks that will be done on the xmltree.

        .. warning::
            This method is only supported for input versions before the Max5 release

        :param count: number of k-points
        :param gamma: bool that controls if the gamma-point should be included
                      in the k-point mesh
        """
        self._validate_arguments('set_nkpts', args, kwargs)
        self._tasks.append(ModifierTask('set_nkpts', args, kwargs))

    def set_kpath(self, *args: Any, **kwargs: Any) -> None:
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_kpath()` to
        the list of tasks that will be done on the xmltree.

        .. warning::
            This method is only supported for input versions before the Max5 release

        :param kpath: a dictionary with kpoint name as key and k point coordinate as value
        :param count: number of k-points
        :param gamma: bool that controls if the gamma-point should be included
                      in the k-point mesh
        """
        self._validate_arguments('set_kpath', args, kwargs)
        self._tasks.append(ModifierTask('set_kpath', args, kwargs))



"""
Click parameters for easily selecting multiple elements from a list via indices
"""
from __future__ import annotations

from typing import Generic, Sequence, TypeVar, Any
import click

T = TypeVar('T')
"""
Generic Type variable
"""


class IntegerSlice(click.ParamType):
    """
    Click parameter for specifying a range of numbers
    """
    name = 'integer-range'

    def convert(self, value: Any, param: click.Parameter | None, ctx: click.Context | None) -> int | slice:
        if isinstance(value, int):
            return value

        try:
            if '-' in value:
                start, stop = value.split('-')
                return slice(int(start), int(stop) + 1)
            return int(value)
        except TypeError as exc:
            raise click.BadParameter(
                'Please provide either an integer number or two integer numbers separated by -') from exc


class ListElement(IntegerSlice, Generic[T]):
    """
    Click parameter for choosing an (or multiple) element(s) from a list
    """

    def __init__(self, data: Sequence[T], return_list: bool = False) -> None:
        self.data = data
        self.return_list = return_list
        super().__init__()

    def convert(  #type:ignore[override]
            self, value: Any, param: click.Parameter | None, ctx: click.Context | None) -> T | Sequence[T]:
        indices = super().convert(value, param, ctx)

        try:
            values = self.data[indices]
            if not isinstance(values, list) and self.return_list:
                values = [values]  #type:ignore
        except IndexError as exc:
            raise click.BadParameter(
                f'Please provide indices in the range of a list with length {len(self.data)}') from exc
        return values



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains custom conversion functions for the outxml_parser, which
cannot be handled by the standard parsing framework
"""
from __future__ import annotations

from datetime import date
import numpy as np
from masci_tools.util.constants import HTR_TO_EV
from masci_tools.io.common_functions import convert_to_pystd, abs_to_rel, abs_to_rel_f
from typing import Any
from logging import Logger

from . import conversion_function


@conversion_function
def convert_htr_to_ev(out_dict: dict[str, Any],
                      name: str,
                      converted_name: str | None = None,
                      pop: bool = False,
                      add_unit: bool = True,
                      logger: Logger | None = None) -> dict[str, Any]:
    """
    Convert value from htr to eV
    """

    if pop:
        old = out_dict.pop(name, None)
    else:
        old = out_dict.get(name)

    if converted_name is None:
        converted_name = f'{old}_ev'

    if add_unit:
        out_dict[f'{converted_name}_units'] = 'eV'

    if old is None:
        if name in out_dict:
            if logger is not None:
                logger.warning(f'convert_htr_to_ev cannot convert None to eV (Name: {name})')
            out_dict[converted_name] = None
        return out_dict

    old = old[-1]
    if old is not None:
        out_dict.setdefault(converted_name, []).append(old * HTR_TO_EV)
    else:
        if logger is not None:
            logger.warning(f'convert_htr_to_ev cannot convert None to eV (Name: {name})')
        out_dict.setdefault(converted_name, []).append(None)

    return out_dict


@conversion_function
def calculate_total_magnetic_moment(out_dict: dict[str, Any], logger: Logger) -> dict[str, Any]:
    """
    Calculate the the total magnetic moment per cell

    :param out_dict: dict with the already parsed information
    """
    total_charge = out_dict.get('spin_dependent_charge_total', None)

    if total_charge is None:
        if logger is not None:
            logger.warning('calculate_total_magnetic_moment got None')
        return out_dict

    total_charge = total_charge[-1]
    if isinstance(total_charge, list):
        if 'total_magnetic_moment_cell' not in out_dict:
            out_dict['total_magnetic_moment_cell'] = []

        out_dict['total_magnetic_moment_cell'].append(convert_to_pystd(np.abs(total_charge[0] - total_charge[1])))

    return out_dict


@conversion_function
def calculate_walltime(out_dict: dict[str, Any], logger: Logger) -> dict[str, Any]:
    """
    Calculate the walltime from start and end time

    :param out_dict: dict with the already parsed information
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    """

    if out_dict['start_date']['time'] is not None:
        starttimes = out_dict['start_date']['time'].split(':')
    else:
        starttimes = [0, 0, 0]
        msg = 'Starttime was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    if out_dict['end_date']['time'] is not None:
        endtimes = out_dict['end_date']['time'].split(':')
    else:
        endtimes = [0, 0, 0]
        msg = 'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    if out_dict['start_date']['date'] is not None:
        start_date = out_dict['start_date']['date']
    else:
        start_date = None
        msg = 'Startdate was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    if out_dict['end_date']['date'] is not None:
        end_date = out_dict['end_date']['date']
    else:
        end_date = None
        msg = 'Enddate was unparsed, inp.xml prob not complete, do not believe the walltime!'
        if logger is not None:
            logger.warning(msg)

    offset = 0
    if start_date is not None and end_date is not None:
        if start_date != end_date:
            date_sl = [int(ent) for ent in start_date.split('/')]
            date_el = [int(ent) for ent in end_date.split('/')]
            date_s = date(*date_sl)
            date_e = date(*date_el)
            diff = date_e - date_s
            offset = diff.days * 86400

    time = offset + (int(endtimes[0]) - int(starttimes[0])) * 60 * 60 + (
        int(endtimes[1]) - int(starttimes[1])) * 60 + int(endtimes[2]) - int(starttimes[2])
    out_dict['walltime'] = time
    out_dict['walltime_units'] = 'seconds'

    return out_dict


@conversion_function
def convert_ldau_definitions(out_dict: dict[str, Any], logger: Logger) -> dict[str, Any]:
    """
    Convert the parsed information from LDA+U into a more readable dict

    ldau_info has keys for each species with LDA+U ({species_name}/{atom_number})
    and this in turn contains a dict with the LDA+U definition for the given orbital (spdf)

    :param out_dict: dict with the already parsed information
    """
    parsed_ldau = out_dict['ldau_info'].pop('parsed_ldau')
    ldau_species = out_dict['ldau_info'].pop('ldau_species')

    if isinstance(ldau_species['name'], str):
        ldau_species = {key: [val] for key, val in ldau_species.items()}

    if isinstance(parsed_ldau['l'], int):
        parsed_ldau = {key: [val] for key, val in parsed_ldau.items()}

    ldau_definitions = zip(ldau_species['name'], ldau_species['atomic_number'], parsed_ldau['l'])
    for index, ldau_def in enumerate(ldau_definitions):

        species_name, atom_number, orbital = ldau_def

        species_key = f'{species_name}/{atom_number}'
        orbital_key = 'spdf'[orbital]

        ldau_dict = out_dict['ldau_info'].setdefault(species_key, {})
        ldau_dict[orbital_key] = {
            'u': parsed_ldau['u'][index],
            'j': parsed_ldau['j'][index],
            'unit': 'eV',
            'double_counting': 'AMF' if parsed_ldau['l_amf'][index] else 'FLL'
        }

    return out_dict


@conversion_function
def convert_ldahia_definitions(out_dict: dict[str, Any], logger: Logger) -> dict[str, Any]:
    """
    Convert the parsed information from LDA+U into a more readable dict

    ldau_info has keys for each species with LDA+U ({species_name}/{atom_number})
    and this in turn contains a dict with the LDA+U definition for the given orbital (spdf)

    :param out_dict: dict with the already parsed information
    """
    parsed_ldahia = out_dict['ldahia_info'].pop('parsed_ldahia')
    ldahia_species = out_dict['ldahia_info'].pop('ldahia_species')

    if isinstance(ldahia_species['name'], str):
        ldahia_species = {key: [val] for key, val in ldahia_species.items()}

    if isinstance(parsed_ldahia['l'], int):
        parsed_ldahia = {key: [val] for key, val in parsed_ldahia.items()}

    ldau_definitions = zip(ldahia_species['name'], ldahia_species['atomic_number'], parsed_ldahia['l'])
    for index, ldahia_def in enumerate(ldau_definitions):

        species_name, atom_number, orbital = ldahia_def

        species_key = f'{species_name}/{atom_number}'
        orbital_key = 'spdf'[orbital]

        ldahia_dict = out_dict['ldahia_info'].setdefault(species_key, {})

        ldahia_dict[orbital_key] = {
            'u': parsed_ldahia['u'][index],
            'j': parsed_ldahia['j'][index],
            'unit': 'eV',
            'double_counting': 'AMF' if parsed_ldahia['l_amf'][index] else 'FLL',
            'initial_occupation': parsed_ldahia['init_occ'][index]
        }

    return out_dict


@conversion_function
def convert_relax_info(out_dict: dict[str, Any], logger: Logger) -> dict[str, Any]:
    """
    Convert the general relaxation information

    :param out_dict: dict with the already parsed information
    """

    atoms, cell, pbc = out_dict.pop('parsed_relax_info')
    film = not all(pbc)

    positions = [convert_to_pystd(site.position) for site in atoms]
    if film:
        positions = [abs_to_rel_f(position, cell, pbc) for position in positions]
    else:
        positions = [abs_to_rel(position, cell) for position in positions]

    out_dict['relax_brav_vectors'] = cell.tolist()
    out_dict['relax_atom_positions'] = [[round(x, 16) for x in pos] for pos in positions]
    out_dict['relax_atomtype_info'] = [(site.kind, site.symbol) for site in atoms]

    return out_dict


@conversion_function
def convert_forces(out_dict: dict[str, Any], logger: Logger) -> dict[str, Any]:
    """
    Convert the parsed forces from a iteration

    :param out_dict: dict with the already parsed information
    """
    parsed_forces = out_dict.pop('parsed_forces')

    if 'force_largest_component' not in out_dict:
        out_dict['force_largest_component'] = []
        out_dict['force_atoms'] = []
        out_dict['abspos_atoms'] = []

    if isinstance(parsed_forces['atom_type'], int):
        parsed_forces = {key: [val] for key, val in parsed_forces.items()}

    largest_force = 0.0
    forces = []
    abspos = []
    for index, atomType in enumerate(parsed_forces['atom_type']):

        force_x = parsed_forces['f_x'][index]
        force_y = parsed_forces['f_y'][index]
        force_z = parsed_forces['f_z'][index]

        x = parsed_forces['x'][index]
        y = parsed_forces['y'][index]
        z = parsed_forces['z'][index]

        forces.append((atomType, [force_x, force_y, force_z]))
        abspos.append((atomType, [x, y, z]))

        if abs(force_x) > largest_force:
            largest_force = abs(force_x)
        if abs(force_y) > largest_force:
            largest_force = abs(force_y)
        if abs(force_z) > largest_force:
            largest_force = abs(force_z)

    out_dict['force_largest_component'].append(largest_force)
    out_dict['force_atoms'].append(forces)
    out_dict['abspos_atoms'].append(abspos)

    return out_dict



"""
Module for utilities for parsing. Should be independent of the actual parsing functions
"""
from __future__ import annotations

from typing import NamedTuple, Any


class Conversion(NamedTuple):
    name: str
    args: tuple[Any, ...] = ()
    kwargs: dict[str, Any] = {}



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains helper functions for extracting information easily from the
schema_dicts defined for the Fleur input/output

Also provides convenient functions to use just a attribute name for extracting the
attribute from the right place in the given etree
"""
from __future__ import annotations

from masci_tools.io.parsers import fleur_schema
from masci_tools.util.xml.common_functions import add_tag, check_complex_xpath
from masci_tools.util.xml.xpathbuilder import XPathBuilder, FilterType
from masci_tools.util.typing import XPathLike, XMLLike
from lxml import etree
from logging import Logger
import warnings
import copy
import os
from typing import Iterable, Any, overload
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  #type:ignore


def read_constants(root: XMLLike | etree.XPathElementEvaluator,
                   schema_dict: fleur_schema.SchemaDict,
                   logger: Logger | None = None) -> dict[str, float]:
    """
    Reads in the constants defined in the inp.xml
    and returns them combined with the predefined constants from
    fleur as a dictionary

    :param root: root of the etree of the inp.xml file
    :param schema_dict: schema_dictionary of the version of the file to read (inp.xml or out.xml)
    :param logger: logger object for logging warnings, errors

    :return: a python dictionary with all defined constants
    """
    from masci_tools.io.fleur_xml import get_constants  #pylint: disable=cyclic-import
    warnings.warn('read_constants is moved to masci_tools.io.fleur_xml', DeprecationWarning)
    return get_constants(root, schema_dict, logger=logger)  #type: ignore[arg-type]


def evaluate_attribute(node: XMLLike | etree.XPathElementEvaluator,
                       schema_dict: fleur_schema.SchemaDict,
                       name: str,
                       constants: dict[str, float] | None = None,
                       logger: Logger | None = None,
                       complex_xpath: XPathLike | None = None,
                       filters: FilterType | None = None,
                       iteration_path: bool = False,
                       **kwargs: Any) -> Any:
    """
    Evaluates the value of the attribute based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param optional: bool, if True and no logger given none or an empty list is returned

    :returns: list or single value, converted in convert_xml_attribute
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.converters import convert_from_xml

    list_return = kwargs.pop('list_return', False)
    optional = kwargs.pop('optional', False)

    attrib_xpath = _select_attrib_xpath(node, schema_dict, name, iteration_path=iteration_path, **kwargs)

    if complex_xpath is None:
        complex_xpath = XPathBuilder(attrib_xpath, filters=filters, strict=True)
    elif filters is not None:
        if not isinstance(complex_xpath, XPathBuilder):
            raise ValueError(
                'Provide only one of filters or complex_xpath (Except when complx_xpath is given as a XPathBuilder)')
        for key, val in filters.items():
            complex_xpath.add_filter(key, val)
    check_complex_xpath(node, attrib_xpath, complex_xpath)

    stringattribute: list[str] = eval_xpath(node, complex_xpath, logger=logger, list_return=True)  #type:ignore

    if len(stringattribute) == 0:
        if logger is None:
            if not optional:
                raise ValueError(f'No values found for attribute {name}')
        else:
            logger.warning('No values found for attribute %s', name)
        if list_return:
            return []
        return None

    converted_value, suc = convert_from_xml(stringattribute,
                                            schema_dict,
                                            name,
                                            text=False,
                                            constants=constants,
                                            logger=logger,
                                            list_return=list_return)

    if not suc:
        if logger is None:
            raise ValueError(f'Failed to evaluate attribute {name}, Got value: {stringattribute}')
        logger.warning('Failed to evaluate attribute %s, Got value: %s', name, stringattribute)

    return converted_value


def evaluate_text(node: XMLLike | etree.XPathElementEvaluator,
                  schema_dict: fleur_schema.SchemaDict,
                  name: str,
                  constants: dict[str, float] | None = None,
                  logger: Logger | None = None,
                  complex_xpath: XPathLike | None = None,
                  iteration_path: bool = False,
                  filters: FilterType | None = None,
                  **kwargs: Any) -> Any:
    """
    Evaluates the text of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param optional: bool, if True and no logger given none or an empty list is returned

    :returns: list or single value, converted in convert_xml_text
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.converters import convert_from_xml

    list_return = kwargs.pop('list_return', False)
    optional = kwargs.pop('optional', False)

    tag_xpath = _select_tag_xpath(node, schema_dict, name, iteration_path=iteration_path, **kwargs)
    if complex_xpath is None:
        complex_xpath = XPathBuilder(tag_xpath, filters=filters, strict=True)
    elif filters is not None:
        if not isinstance(complex_xpath, XPathBuilder):
            raise ValueError(
                'Provide only one of filters or complex_xpath (Except when complx_xpath is given as a XPathBuilder)')
        for key, val in filters.items():
            complex_xpath.add_filter(key, val)

    check_complex_xpath(node, tag_xpath, complex_xpath)

    stringtext: list[str] = eval_xpath(node, add_tag(complex_xpath, 'text()'), logger=logger,
                                       list_return=True)  #type:ignore

    for text in stringtext.copy():
        if text.strip() == '':
            stringtext.remove(text)

    if len(stringtext) == 0:
        if logger is None:
            if not optional:
                raise ValueError(f'No text found for tag {name}')
        else:
            logger.warning('No text found for tag %s', name)
        if list_return:
            return []
        return None

    converted_value, suc = convert_from_xml(stringtext,
                                            schema_dict,
                                            name,
                                            text=True,
                                            constants=constants,
                                            logger=logger,
                                            list_return=list_return)

    if not suc:
        if logger is None:
            raise ValueError(f'Failed to evaluate text for tag {name}, Got text: {stringtext}')
        logger.warning('Failed to evaluate text for tag %s, Got text: %s', name, stringtext)

    return converted_value


def evaluate_tag(node: XMLLike | etree.XPathElementEvaluator,
                 schema_dict: fleur_schema.SchemaDict,
                 name: str,
                 constants: dict[str, float] | None = None,
                 logger: Logger | None = None,
                 subtags: bool = False,
                 text: bool = True,
                 complex_xpath: XPathLike | None = None,
                 iteration_path: bool = False,
                 filters: FilterType | None = None,
                 **kwargs: Any) -> Any:
    """
    Evaluates all attributes of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param subtags: optional bool, if True the subtags of the given tag are evaluated
    :param text: optional bool, if True the text of the tag is also parsed
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found

    :returns: dict, with attribute values converted via convert_xml_attribute
    """
    from masci_tools.util.xml.common_functions import eval_xpath, split_off_tag
    from masci_tools.util.xml.converters import convert_from_xml

    only_required = kwargs.pop('only_required', False)
    strict_missing_error = kwargs.pop('strict_missing_error', False)
    ignore = kwargs.pop('ignore', None)
    list_return = kwargs.pop('list_return', False)

    tag_xpath = _select_tag_xpath(node, schema_dict, name, iteration_path=iteration_path, **kwargs)
    if complex_xpath is None:
        complex_xpath = XPathBuilder(tag_xpath, filters=filters, strict=True)
    elif filters is not None:
        if not isinstance(complex_xpath, XPathBuilder):
            raise ValueError(
                'Provide only one of filters or complex_xpath (Except when complx_xpath is given as a XPathBuilder)')
        for key, val in filters.items():
            complex_xpath.add_filter(key, val)

    check_complex_xpath(node, tag_xpath, complex_xpath)

    try:
        tag_info = _select_tag_info(node, schema_dict, name, iteration_path=iteration_path, **kwargs)
    except ValueError as err:
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from tag {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes') from err
        logger.exception(
            'Failed to evaluate attributes from tag %s: '
            'No attributes to parse either the tag does not '
            'exist or it has no attributes', name)
        return {}

    attribs = tag_info['attribs']
    optional = tag_info['optional_attribs']
    tags = tag_info['simple'] | tag_info['complex']
    optional_tags = tag_info['optional']

    if only_required:
        attribs = attribs.difference(optional)
        tags = tags.difference(optional_tags)

    if ignore:
        attribs = attribs.difference(ignore)
        tags = tags.difference(ignore)

    parse_text = name in schema_dict['text_tags'] and text

    if not attribs and not parse_text and not tags:
        if subtags:
            return {}
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from tag {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes')
        logger.error(
            'Failed to evaluate attributes from tag %s: '
            'No attributes to parse either the tag does not '
            'exist or it has no attributes', name)
    attrib_list = sorted(list(attribs.original_case.values()))

    out_dict: dict[str, Any] = {}

    for attrib in attrib_list:

        stringattribute: list[str] = eval_xpath(node,
                                                add_tag(complex_xpath, f'@{attrib}'),
                                                logger=logger,
                                                list_return=True)  #type:ignore

        if len(stringattribute) == 0:
            if logger is None:
                if strict_missing_error and attrib not in optional:
                    raise ValueError(f'No values found for attribute {attrib} at tag {name}')
            else:
                logger.warning('No values found for attribute %s at tag %s', attrib, name)
            if list_return:
                out_dict[attrib] = []
            else:
                out_dict[attrib] = None
            continue

        out_dict[attrib], suc = convert_from_xml(stringattribute,
                                                 schema_dict,
                                                 attrib,
                                                 text=False,
                                                 constants=constants,
                                                 logger=logger,
                                                 list_return=list_return)
        if not suc:
            if logger is None:
                raise ValueError(f'Failed to evaluate attribute {attrib}, Got value: {stringattribute}')
            logger.warning('Failed to evaluate attribute %s, Got value: %s', attrib, stringattribute)

    if parse_text:

        _, name = split_off_tag(tag_xpath)
        stringtext: list[str] = eval_xpath(node, add_tag(complex_xpath, 'text()'), logger=logger,
                                           list_return=True)  #type:ignore

        for textval in stringtext.copy():
            if textval.strip() == '':
                stringtext.remove(textval)

        if len(stringtext) == 0:
            if logger is None:
                if not optional:
                    raise ValueError(f'No text found for tag {name}')
            else:
                logger.warning('No text found for tag %s', name)
            if list_return:
                out_dict[name] = []
            else:
                out_dict[name] = None

        out_dict[name], suc = convert_from_xml(stringtext,
                                               schema_dict,
                                               name,
                                               text=True,
                                               constants=constants,
                                               logger=logger,
                                               list_return=list_return)

    if subtags:
        for tag in tags:
            if tag in out_dict:
                if logger is None:
                    raise ValueError(f'Conflicting key {tag}: Key is already in the output dictionary')
                logger.error('Conflicting key %s: Key is already in the output dictionary', tag)
            out_dict[tag] = []

        sub_nodes: list[etree._Element] = eval_xpath(node, complex_xpath, logger=logger, list_return=True)  #type:ignore
        for sub_node in sub_nodes:
            for tag in tags:
                if tag_exists(sub_node, schema_dict, tag):
                    out_dict[tag].append(
                        evaluate_tag(sub_node,
                                     schema_dict,
                                     tag,
                                     constants=constants,
                                     logger=logger,
                                     subtags=True,
                                     ignore=ignore,
                                     only_required=only_required,
                                     strict_missing_error=strict_missing_error,
                                     list_return=list_return,
                                     text=text))

        for tag in tags:
            for indx, sub_dict in enumerate(out_dict[tag]):
                if not sub_dict:
                    out_dict[tag].remove(sub_dict)
                elif len(sub_dict) == 1 and tag in sub_dict:
                    out_dict[tag][indx] = sub_dict[tag]

        for tag in tags:
            if len(out_dict[tag]) == 1:
                out_dict[tag] = out_dict[tag][0]
            elif len(out_dict[tag]) == 0:
                out_dict.pop(tag)

    return out_dict


def evaluate_single_value_tag(node: XMLLike | etree.XPathElementEvaluator,
                              schema_dict: fleur_schema.SchemaDict,
                              name: str,
                              constants: dict[str, float] | None = None,
                              logger: Logger | None = None,
                              complex_xpath: XPathLike | None = None,
                              **kwargs: Any) -> Any:
    """
    Evaluates the value and unit attribute of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                               the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    :returns: value and unit, both converted in convert_xml_attribute
    """

    only_required = kwargs.get('only_required', False)
    ignore = kwargs.pop('ignore', ['comment'])

    value_dict = evaluate_tag(node,
                              schema_dict,
                              name,
                              constants=constants,
                              logger=logger,
                              complex_xpath=complex_xpath,
                              ignore=ignore,
                              **kwargs)

    if value_dict.get('value') is None:
        if logger is None:
            raise ValueError(f"Failed to evaluate singleValue from tag {name}: Has no 'value' attribute")
        logger.warning("Failed to evaluate singleValue from tag %s: Has no 'value' attribute", name)

    if value_dict.get('units') is None and not only_required and 'units' not in ignore:
        if logger is None:
            raise ValueError(f"Failed to evaluate singleValue from tag {name}: Has no 'units' attribute")
        logger.warning("Failed to evaluate singleValue from tag %s: Has no 'units' attribute", name)

    return value_dict


def evaluate_parent_tag(node: XMLLike | etree.XPathElementEvaluator,
                        schema_dict: fleur_schema.SchemaDict,
                        name: str,
                        constants: dict[str, float] | None = None,
                        logger: Logger | None = None,
                        complex_xpath: XPathLike | None = None,
                        iteration_path: bool = False,
                        filters: FilterType | None = None,
                        **kwargs: Any) -> Any:
    """
    Evaluates all attributes of the parent tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found

    :returns: dict, with attribute values converted via convert_xml_attribute
    """
    from masci_tools.util.xml.common_functions import eval_xpath, get_xml_attribute
    from masci_tools.util.xml.converters import convert_from_xml

    strict_missing_error = kwargs.pop('strict_missing_error', False)
    list_return = kwargs.pop('list_return', False)
    only_required = kwargs.pop('only_required', False)
    ignore = kwargs.pop('ignore', None)

    tag_xpath = _select_tag_xpath(node, schema_dict, name, iteration_path=iteration_path, **kwargs)
    if complex_xpath is None:
        complex_xpath = XPathBuilder(tag_xpath, filters=filters, strict=True)
    elif filters is not None:
        if not isinstance(complex_xpath, XPathBuilder):
            raise ValueError(
                'Provide only one of filters or complex_xpath (Except when complx_xpath is given as a XPathBuilder)')
        for key, val in filters.items():
            complex_xpath.add_filter(key, val)

    check_complex_xpath(node, tag_xpath, complex_xpath)

    #Which attributes are expected
    try:
        tag_info = _select_tag_info(node, schema_dict, name, parent=True, iteration_path=iteration_path, **kwargs)
    except ValueError as err:
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from parent tag of {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes') from err
        logger.exception(
            'Failed to evaluate attributes from parent tag of %s: '
            'No attributes to parse either the tag does not '
            'exist or it has no attributes', name)
        return {}

    attribs = tag_info['attribs']
    optional = tag_info['optional_attribs']

    if only_required:
        attribs = attribs.difference(optional)

    if ignore is not None:
        attribs = attribs.difference(ignore)

    if not attribs:
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from parent tag of {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes')
        logger.error(
            'Failed to evaluate attributes from parent tag of %s: '
            'No attributes to parse either the tag does not '
            'exist or it has no attributes', name)
    attrib_list = sorted(list(attribs.original_case.values()))

    elems: list[etree._Element] = eval_xpath(node, complex_xpath, logger=logger, list_return=True)  #type:ignore

    out_dict: dict[str, Any] = {}
    for attrib in attrib_list:
        out_dict[attrib] = []

    for elem in elems:
        parent = elem.getparent()
        if parent is None:
            if logger is None:
                raise ValueError(f'No parent found tag {name}')
            logger.warning('No parent found tag %s', name)
            continue
        for attrib in attrib_list:

            try:
                stringattribute = get_xml_attribute(parent, attrib, logger=logger)
            except ValueError:
                stringattribute = None

            if stringattribute is None:
                if logger is None:
                    if strict_missing_error and attrib not in optional:
                        raise ValueError(f'No values found for attribute {attrib} for parent tag of {name}')
                else:
                    logger.warning('No values found for attribute %s for parent tag of %s', attrib, name)
                out_dict[attrib].append(None)
                continue

            value, suc = convert_from_xml(stringattribute,
                                          schema_dict,
                                          attrib,
                                          text=attrib in tag_info['text'],
                                          constants=constants,
                                          logger=logger,
                                          list_return=list_return)

            out_dict[attrib].append(value)

            if not suc:
                if logger is None:
                    raise ValueError(f'Failed to evaluate attribute {attrib}, Got value: {stringattribute}')
                logger.warning('Failed to evaluate attribute %s, Got value: %s', attrib, stringattribute)

    if all(len(x) == 1 for x in out_dict.values()) and not list_return:
        out_dict = {key: val[0] for key, val in out_dict.items()}

    return out_dict


def attrib_exists(node: XMLLike | etree.XPathElementEvaluator,
                  schema_dict: fleur_schema.SchemaDict,
                  name: str,
                  logger: Logger | None = None,
                  iteration_path: bool = False,
                  filters: FilterType | None = None,
                  **kwargs: Any) -> bool:
    """
    Evaluates whether the attribute exists in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: bool, True if any tag with the attribute exists
    """
    from masci_tools.util.xml.common_functions import eval_xpath, split_off_attrib

    attrib_xpath = _select_attrib_xpath(node, schema_dict, name, iteration_path=iteration_path, **kwargs)
    tag_xpath, attrib_name = split_off_attrib(attrib_xpath)
    tag_xpath_builder = XPathBuilder(tag_xpath, filters=filters, strict=True)

    tags: list[etree._Element] = eval_xpath(node, tag_xpath_builder, logger=logger, list_return=True)  #type:ignore
    return any(attrib_name in tag.attrib for tag in tags)


def tag_exists(node: XMLLike | etree.XPathElementEvaluator,
               schema_dict: fleur_schema.SchemaDict,
               name: str,
               logger: Logger | None = None,
               **kwargs: Any) -> bool:
    """
    Evaluates whether the tag exists in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    :returns: bool, True if any nodes with the path exist
    """
    return get_number_of_nodes(node, schema_dict, name, logger=logger, **kwargs) != 0


def get_number_of_nodes(node: XMLLike | etree.XPathElementEvaluator,
                        schema_dict: fleur_schema.SchemaDict,
                        name: str,
                        logger: Logger | None = None,
                        **kwargs: Any) -> int:
    """
    Evaluates the number of occurrences of the tag in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
        :param filters: Dict specifying constraints to apply on the xpath.
                        See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    :returns: number of nodes for the given tag
    """
    result = eval_simple_xpath(node, schema_dict, name, logger=logger, list_return=True, **kwargs)
    if not isinstance(result, list):
        raise ValueError(f'Invalid result for length determination: {str(result)}')
    return len(result)


@overload
def eval_simple_xpath(node: XMLLike | etree.XPathElementEvaluator,
                      schema_dict: fleur_schema.SchemaDict,
                      name: str,
                      logger: Logger | None = ...,
                      iteration_path: bool = ...,
                      filters: FilterType | None = ...,
                      list_return: Literal[True] = ...,
                      **kwargs: Any) -> list[etree._Element]:
    ...


@overload
def eval_simple_xpath(node: XMLLike | etree.XPathElementEvaluator,
                      schema_dict: fleur_schema.SchemaDict,
                      name: str,
                      logger: Logger | None = ...,
                      iteration_path: bool = ...,
                      filters: FilterType | None = ...,
                      list_return: Literal[False] = ...,
                      **kwargs: Any) -> etree._Element | list[etree._Element]:
    ...


def eval_simple_xpath(node: XMLLike | etree.XPathElementEvaluator,
                      schema_dict: fleur_schema.SchemaDict,
                      name: str,
                      logger: Logger | None = None,
                      iteration_path: bool = False,
                      filters: FilterType | None = None,
                      list_return: bool = False,
                      **kwargs: Any) -> etree._Element | list[etree._Element]:
    """
    Evaluates a simple xpath expression of the tag in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details
    :param list_return: bool, if True a list is always returned

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: etree Elements obtained via the simple xpath expression
    """
    from masci_tools.util.xml.common_functions import eval_xpath

    tag_xpath = _select_tag_xpath(node, schema_dict, name, iteration_path=iteration_path, **kwargs)
    tag_xpath_builder = XPathBuilder(tag_xpath, strict=True, filters=filters)

    return eval_xpath(node, tag_xpath_builder, logger=logger, list_return=list_return)  #type: ignore[return-value]


def reverse_xinclude(xmltree: etree._ElementTree, schema_dict: fleur_schema.SchemaDict, included_tags: Iterable[str],
                     **kwargs: os.PathLike) -> tuple[etree._ElementTree, dict[os.PathLike | str, etree._ElementTree]]:
    """
    Split the xmltree back up according to the given included tags.
    The original xmltree will be returned with the corresponding xinclude tags
    and the included trees are returned in a dict mapping the inserted filename
    to the extracted tree

    Tags for which no known filename is known are returned under unknown-1.xml, ...
    The following tags have known filenames:

        - `relaxation`: ``relax.xml``
        - `kPointLists`: ``kpts.xml``
        - `symmetryOperations`: ``sym.xml``
        - `atomSpecies`: ``species.xml``
        - `atomGroups`: ``atoms.xml``

    Additional mappings can be given in the keyword arguments

    :param xmltree: an xml-tree which will be processed
    :param schema_dict: Schema dictionary containing all the necessary information
    :param included_tags: Iterable of str, containing the names of the tags to be excluded

    :returns: xmltree with the inseerted xinclude tags and a dict mapping the filenames
              to the excluded trees

    :raises ValueError: if the tag can not be found in the given xmltree
    """
    from masci_tools.util.xml.common_functions import eval_xpath

    INCLUDE_NSMAP = {'xi': 'http://www.w3.org/2001/XInclude'}
    INCLUDE_TAG = etree.QName(INCLUDE_NSMAP['xi'], 'include')
    FALLBACK_TAG = etree.QName(INCLUDE_NSMAP['xi'], 'fallback')

    excluded_tree = copy.deepcopy(xmltree)

    include_file_names: dict[str, os.PathLike | str] = {
        'relaxation': 'relax.xml',
        'kPointLists': 'kpts.xml',
        'symmetryOperations': 'sym.xml',
        'atomSpecies': 'species.xml',
        'atomGroups': 'atoms.xml'
    }

    include_file_names = {**include_file_names, **kwargs}

    unknown_file_names = 0
    included_trees = {}
    root = excluded_tree.getroot()

    if not all(isinstance(tag, str) for tag in included_tags):
        raise ValueError(f'included_tags is not made up of strings: {included_tags}')

    for tag in included_tags:
        if tag in include_file_names:
            file_name = include_file_names[tag]
        else:
            warnings.warn(f'No filename known for tag {tag}')
            unknown_file_names += 1
            file_name = f'unknown-{unknown_file_names}.xml'

        try:
            tag_xpath = schema_dict.tag_xpath(tag)
        except Exception as err:
            raise ValueError(f'Cannot determine place of included tag {tag}') from err
        included_tag_res: list[etree._Element] = eval_xpath(root, tag_xpath, list_return=True)  #type:ignore

        if len(included_tag_res) != 1:
            raise ValueError(f'Cannot determine place of included tag {tag}')
        included_tag = included_tag_res[0]

        included_trees[file_name] = etree.ElementTree(included_tag)

        parent = included_tag.getparent()
        if parent is None:
            raise ValueError('Could not find parent of included tag')

        xinclude_elem = etree.Element(INCLUDE_TAG, href=os.fspath(file_name), nsmap=INCLUDE_NSMAP)
        xinclude_elem.append(etree.Element(FALLBACK_TAG))

        parent.replace(included_tag, xinclude_elem)

    if 'relax.xml' not in included_trees:
        #The relax.xml include should always be there
        xinclude_elem = etree.Element(INCLUDE_TAG, href='relax.xml', nsmap=INCLUDE_NSMAP)
        xinclude_elem.append(etree.Element(FALLBACK_TAG))
        root.append(xinclude_elem)

    etree.indent(excluded_tree)
    for tree in included_trees.values():
        etree.indent(tree)

    return excluded_tree, included_trees


def _select_tag_xpath(node: XMLLike | etree.XPathElementEvaluator,
                      schema_dict: fleur_schema.SchemaDict,
                      name: str,
                      iteration_path: bool = False,
                      **kwargs: Any) -> str:
    """
    Select the simple tag xpath used for the evaluation function in this module
    based on the given node and the specifications

    1. If the node is an xmltree or an element with the tag of a root tag
       the normal tag_xpath method is used
    2. If the node is an etree Element not of an root tag the
       relative_tag_xpath method is used
    3. If iteration_path=True and the schema dict is an OutputSchemaDict
       use the iteration_tag_xpath/relative_iteration_tag_xpath according to
       the rules above

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: str of the tag xpath
    """
    if iteration_path:
        if not isinstance(schema_dict, fleur_schema.OutputSchemaDict):
            raise ValueError('iteration_path=True can only be used with OutputSchemaDict')

    root_tags: tuple[str, ...] = (schema_dict['root_tag'],)
    if isinstance(schema_dict, fleur_schema.OutputSchemaDict):
        root_tags += tuple(schema_dict['iteration_tags'])

    xpath = None
    if isinstance(node, etree._Element):
        if node.tag not in root_tags:
            if iteration_path:
                xpath = schema_dict.relative_iteration_tag_xpath(name, node.tag, **kwargs)  #type:ignore
            else:
                xpath = schema_dict.relative_tag_xpath(name, node.tag, **kwargs)

    if xpath is None:
        if iteration_path:
            xpath = schema_dict.iteration_tag_xpath(name, **kwargs)  #type:ignore
        else:
            xpath = schema_dict.tag_xpath(name, **kwargs)

    return xpath


def _select_attrib_xpath(node: XMLLike | etree.XPathElementEvaluator,
                         schema_dict: fleur_schema.SchemaDict,
                         name: str,
                         iteration_path: bool = False,
                         **kwargs: Any) -> str:
    """
    Select the simple attrib xpath used for the evaluation function in this module
    based on the given node and the specifications

    1. If the node is an xmltree or an element with the tag of a root tag
       the normal attrib_xpath method is used
    2. If the node is an etree Element not of an root tag the
       relative_attrib_xpath method is used
    3. If iteration_path=True and the schema dict is an OutputSchemaDict
       use the iteration_attrib_xpath/relative_iteration_attrib_xpath according to
       the rules above

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: str of the tag xpath
    """
    if iteration_path:
        if not isinstance(schema_dict, fleur_schema.OutputSchemaDict):
            raise ValueError('iteration_path=True can only be used with OutputSchemaDict')

    root_tags: tuple[str, ...] = (schema_dict['root_tag'],)
    if isinstance(schema_dict, fleur_schema.OutputSchemaDict):
        root_tags += tuple(schema_dict['iteration_tags'])

    xpath = None
    if isinstance(node, etree._Element):
        if node.tag not in root_tags:
            if iteration_path:
                xpath = schema_dict.relative_iteration_attrib_xpath(name, node.tag, **kwargs)  #type:ignore
            else:
                xpath = schema_dict.relative_attrib_xpath(name, node.tag, **kwargs)

    if xpath is None:
        if iteration_path:
            xpath = schema_dict.iteration_attrib_xpath(name, **kwargs)  #type:ignore
        else:
            xpath = schema_dict.attrib_xpath(name, **kwargs)

    return xpath


def _select_tag_info(node: XMLLike | etree.XPathElementEvaluator,
                     schema_dict: fleur_schema.SchemaDict,
                     name: str,
                     iteration_path: bool = False,
                     iteration_tag: str = 'iteration',
                     contains: str | Iterable[str] | None = None,
                     **kwargs: Any) -> fleur_schema.schema_dict.TagInfo:
    """
    Get the tag information used for the evaluation function in this module
    based on the given node and the specifications

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param iteration_path: bool if True and the SchemaDict is of an output schema an absolute path into
                           the iteration element is constructed
    :param iteration_tag: name of the iteration tag. Onlt used for iteration_path=True

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    All other Kwargs are passed on to the tag_info method

    :returns: dict with the tag information
    """

    if contains is None:
        contains = set()
    elif isinstance(contains, str):
        contains = {contains}
    else:
        contains = set(contains)

    if iteration_path:
        if not isinstance(schema_dict, fleur_schema.OutputSchemaDict):
            raise ValueError('iteration_path=True can only be used with OutputSchemaDict')

    root_tags: tuple[str, ...] = (schema_dict['root_tag'],)
    if isinstance(schema_dict, fleur_schema.OutputSchemaDict):
        root_tags += tuple(schema_dict['iteration_tags'])

    if isinstance(node, etree._Element):
        if node.tag not in root_tags:
            if iteration_path:
                iteration_xpath = schema_dict.tag_xpath(iteration_tag)
                if f'/{node.tag}' not in iteration_xpath:
                    contains.add(node.tag)
            else:
                contains.add(node.tag)

    return schema_dict.tag_info(name, contains=contains, **kwargs)



"""
Modules with functions for visualizing things related to greensfunctions from fleur
"""
from masci_tools.vis.common import spectral_function


def plot_kresolved_greensfunction(greensfunction, spin=None, backend=None, **kwargs):
    """
    Plot a k-resolved :py:class:`~masci_tools.tools.greensfunction.GreensFunction` calculated
    on a kpoint path as a colormesh plot. The energy contour should be chosen to be a equidistant
    grid

    :param greensfunction: :py:class:`~masci_tools.tools.greensfunction.GreensFunction` to plot
    :param spin: int which spin index to plot

    All other Kwargs are passed on to :py:func:`~masci_tools.vis.plot_methods.plot_spectral_function()`
    """

    if not greensfunction.kresolved:
        raise ValueError("Only kresolved green's functions can be plotted as spectral functions")

    if greensfunction.kpath is None:
        raise ValueError("Only green's functions calculated on a k-path can be plotted as spectral functions")

    special_kpoints = []
    for k_index, label in zip(greensfunction.extras['special_kpoint_indices'],
                              greensfunction.extras['special_kpoint_labels']):
        special_kpoints.append((label, greensfunction.kpath[k_index]))

    if spin is not None:
        spectral_func = greensfunction.trace_energy_dependence(spin=spin)
    else:
        spectral_func = greensfunction.trace_energy_dependence(spin=1)
        if greensfunction.nspins == 2:
            spectral_func += greensfunction.trace_energy_dependence(spin=2)

    return spectral_function(greensfunction.kpath,
                             greensfunction.points.real,
                             spectral_func,
                             special_kpoints=special_kpoints,
                             backend=backend,
                             **kwargs)
